# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import logging
import os
import pickle
import shutil
import tempfile
import zipfile
from email.message import EmailMessage
from pathlib import Path
from typing import Optional, Tuple

import bs4
import platformdirs
import requests


ET_DIR = Path(platformdirs.user_data_dir(appname="acconeer_exptool", appauthor="Acconeer AB"))
CODENAME = "plugoneer"
COOKIE_DIR = ET_DIR / CODENAME / "cookies"
COOKIE_FILEPATH = COOKIE_DIR / "developer_page.pickle"
REQUEST_URL = "https://developer.acconeer.com/log-in/"
REFERER_URL = "https://developer.acconeer.com"
ACC_DEV_AJAX_URL = "https://developer.acconeer.com/wp-admin/admin-ajax.php"
A121_SW_URL = "https://developer.acconeer.com/home/a121-docs-software/"
WORDPRESS_PROFILE_BUILDER_OPTIONS = {
    "rememberme": "forever",
    "wp-submit": "Log+In",
    "wppb_login": "true",
    "wppb_request_url": REQUEST_URL,
    "wppb_redirect_priority": "normal",
    "wppb_referer_url": REFERER_URL,
    "_wp_http_referer": "/log-in/",
    "wppb_redirect_check": "true",
}

BIN_FETCH_PROMPT = (
    "To fetch the latest image you need to log into your Acconeer Developer account.\n\n"
    "If you don't have an account you can create one here:\n"
    "https://developer.acconeer.com/create-account/ \n\n"
    "Lost your password? Recover it here:\n"
    "https://developer.acconeer.com/recover-password/"
)

log = logging.getLogger(__name__)


def get_cookies() -> requests.cookies.RequestsCookieJar:
    if os.path.exists(COOKIE_FILEPATH):
        with open(COOKIE_FILEPATH, "rb") as handle:
            cookie_jar = pickle.load(handle)
        cookie_jar.clear_expired_cookies()
        cookies = cookie_jar.get_dict()
        if not cookies:
            cookie_jar = None
    else:
        cookie_jar = None
    return cookie_jar


def _make_cookie_dir() -> None:
    COOKIE_DIR.mkdir(parents=True, exist_ok=True)


def save_cookies(cookies: requests.cookies.RequestsCookieJar) -> None:
    _make_cookie_dir()
    with open(COOKIE_FILEPATH, "wb") as handle:
        pickle.dump(cookies, handle)
    log.debug("Login saved")


def clear_cookies() -> None:
    if os.path.exists(COOKIE_FILEPATH):
        os.remove(COOKIE_FILEPATH)
        log.debug("Login session removed")
    else:
        log.debug("No previous session saved")


def login(email: str, password: str) -> Optional[requests.cookies.RequestsCookieJar]:
    with requests.Session() as session:
        login_page_get_resp = session.get(REQUEST_URL)
        soup = bs4.BeautifulSoup(login_page_get_resp.content, "html.parser")
        csrf_input_elem = soup.find("input", {"id": "CSRFToken-wppb"})
        assert isinstance(csrf_input_elem, bs4.Tag)
        token = csrf_input_elem.get("value")

        credentials = {
            "log": email,
            "pwd": password,
            "CSRFToken-wppb": token,
        }
        login_post_resp = session.post(
            REQUEST_URL,
            data={**credentials, **WORDPRESS_PROFILE_BUILDER_OPTIONS},
            allow_redirects=False,
        )

        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Location
        redirection_url = login_post_resp.headers["Location"]

        if "loginerror" in redirection_url:
            log.warning("Wrong email or password")
            return None

        cookies = login_post_resp.cookies
        if len(cookies) == 0:
            return None

        return cookies


def get_content(
    cookies: requests.cookies.RequestsCookieJar,
) -> Tuple[bool, requests.Session, requests.Response]:
    with requests.Session() as session:
        response = session.get(url=REFERER_URL, cookies=cookies.get_dict())

        if response.url == REQUEST_URL or response.headers.get("Expires") is None:
            return False, session, response
        elif response.headers.get("Expires") is not None:
            return True, session, response
        else:
            raise Exception


def is_redirected_to_login_page(response: requests.Response) -> bool:
    return bool(response.url == REQUEST_URL)


def download(
    session: requests.Session,
    cookies: requests.cookies.RequestsCookieJar,
    path: str,
    device: str,
) -> Tuple[str, str]:
    device = device.lower()

    page_device_map = {"xc120": "xe121"}

    page_device = page_device_map.get(device, device)

    response = session.get(url=A121_SW_URL)

    soup = bs4.BeautifulSoup(response.content, "html.parser")
    link = soup.findAll("a", {"href": lambda l: l and page_device in l})

    if not link:
        msg = f"No download found for device '{device}'"
        raise Exception(msg)

    response = session.get(url=link[0]["href"])
    soup = bs4.BeautifulSoup(response.content, "html.parser")

    zip_url = soup.findAll(
        "a", {"href": lambda l: l and device in l and "exploration_server" in l}
    )
    if not zip_url:
        msg = f"No image found for device '{device}'"
        raise Exception(msg)
    zip_url = zip_url[0]["href"]

    log.debug("File to download: {}".format(zip_url))

    tc = "tc_accepted=1&tc_submit=Download"
    headers = {"Content-Type": "application/x-www-form-urlencoded"}

    r = session.post(zip_url, data=tc, allow_redirects=False, cookies=cookies, headers=headers)

    email_msg = EmailMessage()
    content_disposition = r.headers["Content-Disposition"]
    # EmailMessage.__setitem__ does magic things here
    email_msg["Content-Disposition"] = content_disposition
    filename = email_msg.get_filename()
    if filename is None:
        msg = f"Response from POST to {zip_url} did not contain a filename."
        raise Exception(msg)

    version = _get_version(filename)

    filepath = os.path.join(path, filename)

    with open(filepath, "wb") as output_file:
        output_file.write(r.content)
        log.debug("Download completed at {}".format(filepath))

    log.debug("Extracting files..")

    with tempfile.TemporaryDirectory() as tmp_dir:
        bin_found = False
        with zipfile.ZipFile(filepath, "r") as zipObject:
            listOfFileNames = zipObject.namelist()
            for fileName in listOfFileNames:
                if fileName.endswith(".bin"):
                    if bin_found:
                        msg = "Multiple images in directory"
                        raise Exception(msg)
                    bin_path = zipObject.extract(fileName, os.path.join(path, tmp_dir))
                    bin_found = True
            if bin_found is False:
                msg = "No images found in directory"
                raise Exception(msg)

        dst_path = os.path.join(path, os.path.basename(bin_path))
        shutil.move(bin_path, dst_path)

        os.remove(filepath)

    return dst_path, version


def _get_version(bin_file: Optional[str]) -> str:
    version = "n/a"

    if bin_file is not None:
        version = Path(bin_file).stem

    return version



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from typing_extensions import Protocol

from acconeer.exptool.app.new.backend import GeneralMessage


class PlotPluginInterface(Protocol):
    def handle_message(self, message: GeneralMessage) -> None: ...

    def draw(self) -> None: ...



# Copyright (c) Acconeer AB, 2024-2025
# All rights reserved

from __future__ import annotations

import copy
from typing import List

import attrs

from acconeer.exptool import a121

from ._aggregator import ProcessorSpec
from ._processors import MeasurementType, ProcessorMode


def get_calibrate_offset_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        start_point=-30,
        num_points=50,
        step_length=1,
        profile=a121.Profile.PROFILE_1,
        hwaas=64,
        sweeps_per_frame=1,
        enable_loopback=True,
        phase_enhancement=True,
        iq_imbalance_compensation=True,
    )


def get_calibrate_noise_session_config(
    session_config: a121.SessionConfig, sensor_ids: List[int]
) -> a121.SessionConfig:
    noise_session_config = copy.deepcopy(session_config)

    for sensor_id in sensor_ids:
        for group in noise_session_config.groups:
            group[sensor_id].sweeps_per_frame = 1
            # Set num_points to a high number to get sufficient number of data points to
            # estimate the standard deviation. Extra num_points for step_length = 1 together
            # with profile = 5 due to filter margin and cropping
            if any(
                ss.step_length == 1 and ss.profile == a121.Profile.PROFILE_5
                for ss in group[sensor_id].subsweeps
            ):
                num_points = 352
            else:
                num_points = 220
            for subsweep in group[sensor_id].subsweeps:
                subsweep.enable_tx = False
                subsweep.step_length = 1
                subsweep.start_point = 0
                subsweep.num_points = num_points

    return noise_session_config


def update_processor_mode(
    processor_specs: list[ProcessorSpec], processor_mode: ProcessorMode
) -> list[ProcessorSpec]:
    updated_specs = []
    for spec in processor_specs:
        new_processor_config = attrs.evolve(spec.processor_config, processor_mode=processor_mode)
        updated_specs.append(attrs.evolve(spec, processor_config=new_processor_config))
    return updated_specs


def filter_close_range_spec(specs: list[ProcessorSpec]) -> list[ProcessorSpec]:
    NUM_CLOSE_RANGE_SPECS = 1
    close_range_specs = []
    for spec in specs:
        if spec.processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            close_range_specs.append(spec)
    if len(close_range_specs) != NUM_CLOSE_RANGE_SPECS:
        msg = "Incorrect subsweep config for close range measurement"
        raise ValueError(msg)

    return close_range_specs



# Copyright (c) Acconeer AB, 2024-2025
# All rights reserved

"""
type_migration

This is a module that enables structural handling of value object migration.

Since a lot of our configs/context are subject to updates and recordings
are saved (and not modified), it's very troublesome to do a best effort to
keep backwards compatability.

This module aims to support us in keeping backwards compatibility in all cases
where it's possible.

This module works on the following premises:

1. Necessary parts of old config classes are kept around
2. Each "identity" that is e.g. a Config has an accompanying timeline that
   glues the historical record of e.g. a Config's definitions through time.

For example;

Say we have a simple Config

>>> import attrs
>>> import json
>>> import typing_extensions as te
>>>
>>> def typical_from_json(clz, json_str):
...     return clz(**json.loads(json_str))
...
>>> @attrs.mutable
... class Config_v0:
...     param: float
...     param_enabled: bool
...
...     # below is just to save some vertical space, it's equivalent to @classmethod def from_json ...
...     from_json = classmethod(typical_from_json)
...
>>> Config_v0.from_json('{"param": 0.5, "param_enabled": true}')
Config_v0(param=0.5, param_enabled=True)

But then we realize that our 'param' could be used in 3 different ways, so
we want to replace 'param_enabled' with an Enum, giving us the new version

>>> Mode = te.Literal["A", "B", "C"]  # poor man's Enum
...
>>> @attrs.mutable
... class Config_v1:
...     param: float
...     mode: Mode
...
...     from_json = classmethod(typical_from_json)
...

In order to keep backward compatability, we need to make sure that we can handle
both 'Config_v1' AND 'Config_v0', while our algorithm logic works with 'Config_v1'.

This is where this module comes in. We can use it to define a timeline of the 'Config'
"identity" like so:

>>> from acconeer.exptool import type_migration as tm
>>>
>>> def v0_to_v1(config: Config_v0):
...     # some arbitrary logic
...     mode = (
...         "A" if config.param < 0 else
...         "B" if config.param_enabled else
...         "C"
...     )
...     return Config_v1(param=config.param, mode=mode)
>>>
>>> config_timeline = (
...     tm
...     # declare first epoch of timeline
...     .start(Config_v0)
...
...     # declare how a str could be transformed to the epoch above
...     # and what is expected to fail (if given "wrong" input).
...     #   note that a single epoch can have multiple 'load's.
...     .load(str, Config_v0.from_json, fail=[TypeError])
...
...     # does nothing, but serves as a separator.
...     .nop()
...
...     # defines a new epoch by giving a new type, how to go from
...     # the epoch above to this, new, epoch (and what is expected to fail).
...     .epoch(Config_v1, v0_to_v1, fail=[])
...
...     # same as above
...     .load(str, Config_v1.from_json, fail=[TypeError])
... )

This timeline can now be used to migrate the following values to a 'Config_v1' object:
- A 'Config_v1' object,
- A 'Config_v1' json string,
- A 'Config_v0' object,
- A 'Config_v0' json string.

>>> config_timeline.migrate(Config_v1(param=0.5, mode='A'))
Config_v1(param=0.5, mode='A')

>>> config_timeline.migrate('{"param": 0.5, "mode": "B"}')
Config_v1(param=0.5, mode='B')

>>> config_timeline.migrate(Config_v0(param=0.5, param_enabled=False))
Config_v1(param=0.5, mode='C')

>>> config_timeline.migrate('{"param": 0.5, "param_enabled": true}')
Config_v1(param=0.5, mode='B')

The module also supports requiring context when defining the timeline
(with 'contextual_load' & 'contextual_epoch') and injecting context
to 'migrate' via a Completer (i.e. a function with a specific signature).
"""

from .core import Completer, MigrationError, MigrationErrorGroup, Node, start



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from acconeer.exptool import a121
from acconeer.exptool.a121.algo.parking._ref_app import RefAppConfig


def get_ground_config() -> RefAppConfig:
    ret = RefAppConfig(
        range_start_m=0.1,
        range_end_m=0.4,
        hwaas=24,
        profile=a121.Profile.PROFILE_1,
        update_rate=0.1,
        queue_length_n=3,
        amplitude_threshold=8.0,
        weighted_distance_threshold_m=0.1,
        obstruction_detection=True,
        obstruction_start_m=0.03,
        obstruction_end_m=0.05,
        obstruction_distance_threshold=0.06,
    )
    return ret


def get_pole_config() -> RefAppConfig:
    ret = RefAppConfig(
        range_start_m=0.2,
        range_end_m=3.0,
        hwaas=24,
        profile=a121.Profile.PROFILE_2,
        amplitude_threshold=6.0,
        weighted_distance_threshold_m=0.9,
        update_rate=6.0,
        queue_length_n=20,
        obstruction_detection=False,
    )
    return ret



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import abc
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Generic, Iterator, Optional, TypeVar

import h5py

from acconeer.exptool.app.new import PluginGeneration
from acconeer.exptool.app.new.storage import get_config_dir

from ._message import (
    BackendPluginStateMessage,
    Message,
    StatusFileAccessMessage,
    StatusMessage,
    TimingMessage,
)
from ._tasks import is_task


StateT = TypeVar("StateT")


class BackendPlugin(abc.ABC, Generic[StateT]):
    shared_state: StateT

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        self.callback = callback
        self.key = key
        self.generation = generation

    @contextmanager
    def h5_cache_file(self, write: bool = False) -> Iterator[h5py.File]:
        file_path = (get_config_dir() / "plugin" / self.generation.value / self.key).with_suffix(
            ".h5"
        )

        if write:
            file_path.parent.mkdir(parents=True, exist_ok=True)

        file_mode = "w" if write else "r"
        h5_file = h5py.File(file_path, file_mode)
        yield h5_file
        h5_file.close()

    @contextmanager
    def report_timing(self, name: str) -> Iterator[None]:
        start = time.perf_counter()
        yield
        end = time.perf_counter()
        self.callback(TimingMessage(name=name, start=start, end=end))

    @is_task
    @abc.abstractmethod
    def load_from_cache(self) -> None:
        pass

    @is_task
    @abc.abstractmethod
    def load_from_file(
        self,
        *,
        path: Path,
        config_override: Optional[Any] = None,
        context_override: Optional[Any] = None,
    ) -> None:
        pass

    @is_task
    @abc.abstractmethod
    def set_preset(self, preset_id: int) -> None:
        pass

    @abc.abstractmethod
    def idle(self) -> bool:
        pass

    @abc.abstractmethod
    def attach_client(self, *, client: Any) -> None:
        pass

    @abc.abstractmethod
    def detach_client(self) -> None:
        pass

    @abc.abstractmethod
    def teardown(self) -> None:
        pass

    def broadcast(self) -> None:
        self.callback(BackendPluginStateMessage(state=self.shared_state))

    def send_status_file_access_message(self, file_path: str, opened: bool) -> None:
        self.callback(StatusFileAccessMessage(file_path=file_path, opened=opened))

    def send_status_message(self, message: Optional[str]) -> None:
        self.callback(StatusMessage(status=message))



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from acconeer.exptool.a111.algo import ModuleFamily, ModuleInfo

from ._processor import ProcessingConfiguration, Processor, get_sensor_config
from .ui import PGUpdater


module_info = ModuleInfo(
    key="envelope",
    label="Envelope",
    pg_updater=PGUpdater,
    processing_config_class=ProcessingConfiguration,
    module_family=ModuleFamily.SERVICE,
    sensor_config_class=get_sensor_config,
    processor=Processor,
    multi_sensor=True,
    docs_url="https://docs.acconeer.com/en/latest/services/envelope.html",
)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import itertools
import operator
import typing as t

import attrs
import typing_extensions as te

from PySide6.QtCore import Qt, QTimer
from PySide6.QtWidgets import (
    QGraphicsSceneMouseEvent,
    QGridLayout,
    QTabWidget,
    QToolTip,
    QWidget,
)

import pyqtgraph as pg

from acconeer.exptool import a121
from acconeer.exptool.a121.model import power
from acconeer.exptool.app.new.ui.icons import WARNING_YELLOW
from acconeer.exptool.app.new.ui.resource_tab.event_system import (
    ChangeIdEvent,
    EventBroker,
    IdentifiedServiceUninstalledEvent,
)

from .distance_config_input import DistanceConfigEvent
from .presence_config_input import PresenceConfigEvent
from .session_config_input import SessionConfigEvent


_SECONDS_IF_RATE_UNSET = 2e-2
_DURATION_SB_DECIMALS = 4
_DURATION_SB_RANGE = (1e-4, 100)
_DURATION_SB_STEP = 5e-3

_X_AXIS_SPINBOX_LABEL = "X-axis length"
_MU = "\u00b5"

_A_to_mA = _s_to_ms = _mA_to_uA = 1000

_mA = _ms = 1e-3


class PowerCurveBarGraphItem(pg.BarGraphItem):
    def __init__(self, power_profile: power.CompositeRegion) -> None:
        durations = [r.duration for r in power_profile.flat_iter()]
        starts = list(itertools.accumulate([0] + durations[:-1], operator.add))
        currents = [p.average_current for p in power_profile.flat_iter()]
        colors = [self.tag_color(p.tag) for p in power_profile.flat_iter()]

        super().__init__(
            x0=starts,
            width=durations,
            height=currents,
            brushes=colors,
        )
        self.setAcceptHoverEvents(True)

        self._power_profile = power_profile
        self._ends = [start + duration for start, duration in zip(starts, durations)]

    @staticmethod
    def tag_color(tag: t.Optional[power.EnergyRegion.Tag]) -> str:
        if tag == power.EnergyRegion.Tag.MEASURE:
            return "cornflowerblue"
        if tag == power.EnergyRegion.Tag.OVERHEAD:
            return "red"
        if tag == power.EnergyRegion.Tag.CALIBRATION:
            return "orange"
        if tag == power.EnergyRegion.Tag.IDLE:
            return "forestgreen"
        if tag == power.EnergyRegion.Tag.COMMUNICATION:
            return "goldenrod"

        return "magenta"

    def _profile_description_at_point(self, x: float, y: float) -> str:
        for end, simple_profile in zip(self._ends, self._power_profile.flat_iter()):
            if x >= end:
                continue

            if y < simple_profile.current:
                return "\n".join(
                    [
                        simple_profile.description,
                        "",
                        "Current:",
                        f"    {simple_profile.current * _A_to_mA:.0f} mA",
                        "Duration:",
                        f"    {simple_profile.duration * _s_to_ms:.1f} ms",
                    ]
                )
            else:
                return ""

        return ""

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        x = event.pos().x()
        y = event.pos().y()
        global_pos = event.screenPos()
        tooltip = self._profile_description_at_point(x, y)

        # Work-around to be able to get tooltip-on-click and zoom with rectangle
        QTimer.singleShot(10, lambda: QToolTip.showText(global_pos, tooltip))
        super().mousePressEvent(event)

    def hoverEvent(self, event: t.Any) -> None:
        try:
            x = event.pos().x()
            y = event.pos().y()
        except Exception:
            return

        tooltip = self._profile_description_at_point(x, y)
        self.setToolTip(tooltip)


class _DummyBarGraphItem(pg.BarGraphItem):
    def __init__(self, color: str) -> None:
        super().__init__(brush=color, x0=1, width=1, y0=1, height=1)


class PowerCurvePlotItem(pg.PlotItem):
    """
    Just like pg.PlotItem, but the 'auto' range
    (which is applied when clicking the "A" button) is customizable in X-axis.
    """

    def __init__(self, *args: t.Any, auto_end_x: float, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._auto_end_x: float = auto_end_x

    def set_auto_end_x(self, x: float) -> None:
        self._auto_end_x = x

    def autoBtnClicked(self) -> None:
        if self.autoBtn.mode == "auto":
            # default behaviour is self.enableAutoRange()
            self.setXRange(0, self._auto_end_x)
            self.autoBtn.hide()
        else:
            self.disableAutoRange()


class _EnergyRegionPlot(QWidget):
    @attrs.frozen
    class _State:
        """Whenever this changes, the plot is redrawn (with the new information)"""

        session_config: a121.SessionConfig
        """This received from events"""

        lower_idle_state: t.Optional[power.Sensor.LowerIdleState]

    _MIN_DURATION_S = 20 * _ms
    _MAX_DURATION_S = 10

    def __init__(
        self,
        session_config: a121.SessionConfig,
        lower_idle_state: t.Optional[power.Sensor.LowerIdleState],
        algorithm: power.algo.Algorithm,
    ) -> None:
        super().__init__()

        self._state = self._State(session_config, lower_idle_state)
        self._algorithm: te.Final[power.algo.Algorithm] = algorithm

        duration_of_2_actives = power.session(
            session_config,
            lower_idle_state=lower_idle_state,
            num_actives=2,
            algorithm=algorithm,
        ).duration
        self._plot_item = PowerCurvePlotItem(auto_end_x=duration_of_2_actives)
        self._plot_item.autoBtnClicked()

        self._plot_item.setLabel("left", "Sensor + XM125", units="A")
        self._plot_item.setLabel("bottom", "Duration", units="s")
        self._plot_item.setContentsMargins(0, 0, 0, 10)

        self._plot_widget = pg.PlotWidget(plotItem=self._plot_item)
        self._plot_widget.getViewBox().setMouseMode(pg.ViewBox.PanMode)
        y_min = -0.01
        y_max = 0.1
        self._plot_widget.getViewBox().setLimits(
            yMin=y_min, yMax=y_max, minYRange=y_max - y_min, maxYRange=y_max - y_min
        )

        self._bar_legend = pg.graphicsItems.LegendItem.LegendItem(offset=(75, 10))
        self._bar_legend.setParentItem(self._plot_item)
        self._bar_legend.setEnabled(False)

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._plot_widget, 0, 0, 1, 2)
        self.setLayout(layout)

    def evolve_current_state(self, **kwargs: t.Any) -> None:
        self._state = attrs.evolve(self._state, **kwargs)

    def plot_current_state(self) -> None:
        approx_avg_current = power.converged_average_current(
            self._state.session_config,
            lower_idle_state=self._state.lower_idle_state,
            absolute_tolerance=0.01 * _mA,
            algorithm=self._algorithm,
        )

        if approx_avg_current > 1 * _mA:
            approx_avg_current_formatted = f"{approx_avg_current * _A_to_mA:.0f} mA"
        else:
            milliampere = round(approx_avg_current * _A_to_mA, 2)
            approx_avg_current_formatted = f"{milliampere * _mA_to_uA:.0f} {_MU}A"

        self._plot_widget.clear()
        duration_of_2_actives = power.session(
            self._state.session_config,
            lower_idle_state=self._state.lower_idle_state,
            num_actives=2,
            algorithm=self._algorithm,
        ).duration
        self._plot_item.set_auto_end_x(duration_of_2_actives)

        active_profile = power.group_active(
            self._state.session_config,
            self._state.lower_idle_state,
            algorithm=self._algorithm,
        )

        fifty_first_active_and_idle_profiles = power.session(
            self._state.session_config,
            lower_idle_state=self._state.lower_idle_state,
            num_actives=50,
            algorithm=self._algorithm,
        )

        etc_text = pg.TextItem("and so on ...", anchor=(0, 1))
        etc_text.setPos(fifty_first_active_and_idle_profiles.duration, 0)
        self._plot_widget.addItem(etc_text)

        session_bar_item = PowerCurveBarGraphItem(fifty_first_active_and_idle_profiles)
        self._plot_widget.addItem(session_bar_item)

        power_tag_set = set([p.tag for p in fifty_first_active_and_idle_profiles.flat_iter()])

        self._bar_legend.clear()
        for tag in power_tag_set:
            if tag is not None:
                self._bar_legend.addItem(
                    _DummyBarGraphItem(PowerCurveBarGraphItem.tag_color(tag)), tag.name.title()
                )

        if active_profile.duration > 1.0:
            approx_duration_formatted = f"{active_profile.duration:.1f} s"
        else:
            approx_duration_formatted = f"{active_profile.duration * _s_to_ms:.1f} ms"

        region_bounds = (0, active_profile.duration)

        lri = pg.LinearRegionItem(
            values=region_bounds,
            movable=False,
            pen={"color": "#222", "style": Qt.PenStyle.DashLine, "width": 2},
        )

        ti = pg.TextItem(f"Approx. active duration {approx_duration_formatted}", color="#222")

        self._plot_widget.addItem(lri)
        self._plot_widget.addItem(ti)

        hline_item = pg.InfiniteLine(
            pos=approx_avg_current,
            angle=0,
            pen={"color": "#222", "style": Qt.PenStyle.DashLine, "width": 2},
            label=f"Approx. avg. current ({approx_avg_current_formatted})",
            labelOpts={"color": "#222"},
        )
        self._plot_widget.addItem(hline_item)

        rate = power.configured_rate(self._state.session_config)
        if rate is None:
            return

        if active_profile.duration > 1 / rate:
            rate_warning_text = pg.InfiniteLine(
                pos=0.10,
                angle=0,
                label=f"Cannot keep rate.\nMaximum rate is approx.\n{1 / active_profile.duration:.0f} Hz",
                labelOpts={
                    "color": "#000",
                    "fill": WARNING_YELLOW,
                    "border": "#000",
                },
                pen=pg.mkPen(None),
            )
            self._plot_widget.addItem(rate_warning_text)


class EnergyRegionOutput(QTabWidget):
    INTERESTS: t.ClassVar[set[type]] = {
        SessionConfigEvent,
        IdentifiedServiceUninstalledEvent,
        DistanceConfigEvent,
        PresenceConfigEvent,
        ChangeIdEvent,
    }
    description: t.ClassVar[str] = "\n\n".join(
        [
            "Visualizes a simulated power curve of a confiugration.",
            f"The spinbox {_X_AXIS_SPINBOX_LABEL!r} controls the end of the X-axis.",
            "Click and hold the plotted rectangles to get a description of that rectangle.",
        ]
    )
    window_title = "Power curve"

    def __init__(self, broker: EventBroker) -> None:
        super().__init__()

        self._tabs: dict[str, _EnergyRegionPlot] = {}
        self.setContentsMargins(0, 0, 0, 0)
        self.setStyleSheet("QTabBar { font: bold 14px; font-family: monospace; }")

        broker.install_service(self)
        self.uninstall_function = lambda: broker.uninstall_service(self)
        broker.brief_service(self)

    def handle_event(self, event: t.Any) -> None:
        if isinstance(event, SessionConfigEvent):
            self._handle_session_config_event(event)
        elif isinstance(event, DistanceConfigEvent):
            self._handle_distance_config_event(event)
        elif isinstance(event, PresenceConfigEvent):
            self._handle_presence_config_event(event)
        elif isinstance(event, IdentifiedServiceUninstalledEvent):
            self._handle_identified_service_uninstalled_event(event)
        elif isinstance(event, ChangeIdEvent):
            self._handle_change_id_event(event)
        else:
            raise NotImplementedError

    def _handle_session_config_event(self, event: SessionConfigEvent) -> None:
        if event.service_id not in self._tabs:
            plot_widget = _EnergyRegionPlot(
                event.session_config,
                event.lower_idle_state,
                power.algo.SparseIq(),
            )
            self._tabs[event.service_id] = plot_widget

            self.addTab(plot_widget, event.service_id)
            self.setCurrentWidget(plot_widget)
        else:
            self._tabs[event.service_id].evolve_current_state(
                session_config=event.session_config,
                lower_idle_state=event.lower_idle_state,
            )

        self._tabs[event.service_id].plot_current_state()

    def _handle_distance_config_event(self, event: DistanceConfigEvent) -> None:
        session_config = event.translated_session_config

        if event.service_id not in self._tabs:
            plot_widget = _EnergyRegionPlot(
                session_config,
                event.lower_idle_state,
                power.algo.Distance(),
            )
            self._tabs[event.service_id] = plot_widget

            self.addTab(plot_widget, event.service_id)
            self.setCurrentWidget(plot_widget)
        else:
            self._tabs[event.service_id].evolve_current_state(
                session_config=session_config,
                lower_idle_state=event.lower_idle_state,
            )

        self._tabs[event.service_id].plot_current_state()

    def _handle_presence_config_event(self, event: PresenceConfigEvent) -> None:
        session_config = event.translated_session_config

        if event.service_id not in self._tabs:
            plot_widget = _EnergyRegionPlot(
                session_config,
                event.lower_idle_state,
                power.algo.Presence(),
            )
            self._tabs[event.service_id] = plot_widget

            self.addTab(plot_widget, event.service_id)
            self.setCurrentWidget(plot_widget)
        else:
            self._tabs[event.service_id].evolve_current_state(
                session_config=session_config,
                lower_idle_state=event.lower_idle_state,
            )

        self._tabs[event.service_id].plot_current_state()

    def _handle_identified_service_uninstalled_event(
        self, event: IdentifiedServiceUninstalledEvent
    ) -> None:
        tab_widget = self._tabs.pop(event.id_)
        tab_index = self.indexOf(tab_widget)
        if tab_index != -1:
            self.removeTab(tab_index)

    def _handle_change_id_event(self, event: ChangeIdEvent) -> None:
        tab_widget = self._tabs.pop(event.old_id, None)
        if tab_widget is not None:
            tab_index = self.indexOf(tab_widget)
            self.setTabText(tab_index, event.new_id)
            self._tabs[event.new_id] = tab_widget



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import typing as t

import attrs

from PySide6.QtWidgets import QVBoxLayout

from acconeer.exptool import a121
from acconeer.exptool.a121.model import power
from acconeer.exptool.app.new.ui.components import GroupBox, pidgets
from acconeer.exptool.app.new.ui.components.a121 import SessionConfigEditor
from acconeer.exptool.app.new.ui.resource_tab.event_system import ChangeIdEvent, EventBroker
from acconeer.exptool.app.new.ui.utils import LayoutWrapper, ScrollAreaDecorator, TopAlignDecorator


@attrs.frozen
class SessionConfigEvent:
    service_id: str
    session_config: a121.SessionConfig
    lower_idle_state: t.Optional[power.Sensor.LowerIdleState]


class SessionConfigInput(ScrollAreaDecorator):
    INTERESTS: t.ClassVar[set[type]] = set([ChangeIdEvent])
    description: t.ClassVar[str] = (
        "Specify sensor configuration as in the Stream tab.\n\n"
        + "Additionally, you can specify a lower idle state."
    )
    id_: str = ""

    def __init__(self, broker: EventBroker, initial_config: a121.SessionConfig) -> None:
        layout = QVBoxLayout()
        super().__init__(TopAlignDecorator(LayoutWrapper(layout)))
        self.setMinimumWidth(200)

        self._broker = broker
        self.power_state_selection = pidgets.ComboboxPidgetFactory[
            t.Optional[power.Sensor.LowerIdleState]
        ](
            name_label_text="Lower idle state:",
            name_label_tooltip="The lowest idle states of the sensor are set by the host",
            items=[
                ("Don't use", None),
                ("Hibernate", power.Sensor.IdleState.HIBERNATE),
                ("Off", power.Sensor.IdleState.OFF),
            ],
        ).create(self)

        self.power_state_selection.set_data(None)
        self.power_state_selection.sig_update.connect(self._offer_event_if_config_is_valid)

        self.editor = SessionConfigEditor(supports_multiple_subsweeps=True)
        self.editor._sensor_ids_editor.hide()
        self.editor.set_data(initial_config)
        self.editor.sig_update.connect(self.editor.set_data)
        self.editor.sig_update.connect(self._display_validation_results)
        self.editor.sig_update.connect(self._offer_event_if_config_is_valid)

        wrapped_power_state_selection = GroupBox.vertical(left_header="")
        wrapped_power_state_selection.layout().addWidget(self.power_state_selection)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(wrapped_power_state_selection)
        layout.addWidget(self.editor)

        self.id_ = broker.install_identified_service(self, "sparse-iq")
        self.uninstall_function = lambda: broker.uninstall_identified_service(self, self.id_)
        self._offer_event_if_config_is_valid()
        self.window_title = f"<b><code>[{self.id_}]</code></b> Sparse IQ config"
        self.fixed_title = "Sparse IQ config"

    def _display_validation_results(self) -> None:
        config = self.editor.get_data()

        if config is None:
            return

        self.editor.handle_validation_results(config._collect_validation_results())

    def _offer_event_if_config_is_valid(self) -> None:
        if not self.editor.is_ready:
            return

        config = self.editor.get_data()

        if config is None:
            return

        try:
            config.validate()
        except a121.ValidationResult:
            return

        self._broker.offer_event(
            SessionConfigEvent(
                self.id_,
                config,
                self.power_state_selection.get_data(),
            )
        )

    def handle_event(self, event: t.Any) -> None:
        if isinstance(event, ChangeIdEvent):
            self._handle_change_id_event(event)
        else:
            raise NotImplementedError

    def _handle_change_id_event(self, event: ChangeIdEvent) -> None:
        if event.old_id == self.id_:
            self.id_ = event.new_id



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

import numpy as np

import acconeer.exptool as et


ENVELOPE_BACKGROUND_LEVEL = 100


def get_sensor_config():
    config = et.a111.EnvelopeServiceConfig()
    config.downsampling_factor = 2
    config.range_interval = [0.12, 0.62]
    config.running_average_factor = 0
    config.update_rate = 0.5
    config.hw_accelerated_average_samples = 20
    config.power_save_mode = et.a111.EnvelopeServiceConfig.PowerSaveMode.OFF
    config.asynchronous_measurement = False

    return config


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.session_info = session_info

        self.f = sensor_config.update_rate
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.norm_depths = self.depths / self.depths[-1]  # normalize by max value

        self.update_processing_config(processing_config)

    def update_processing_config(self, processing_config):
        start = self.session_info["range_start_m"]
        step = self.session_info["step_length_m"]

        # Test if we should deal with direct leakage
        if start > processing_config.depth_leak_end:
            self.depth_leak_sample = 0
            self.depth_leak_end = 0
        else:
            self.depth_leak_sample = processing_config.depth_leak_sample
            self.depth_leak_end = processing_config.depth_leak_end

        self.detector_queue_target_length = processing_config.detector_queue_target_length
        self.weight_threshold = processing_config.weight_threshold
        self.weight_ratio_limit = processing_config.weight_ratio_limit
        self.distance_difference_limit = processing_config.distance_difference_limit

        self.leak_sample_index = int(round((self.depth_leak_sample - start) / step))
        self.leak_end_index = int(round((self.depth_leak_end - start) / step))
        self.leak_estimate_depths = np.array(
            [
                self.depths[0] + step * self.leak_sample_index,
                self.depths[0] + step * self.leak_end_index,
            ]
        )
        self.queued_weights = []
        self.queued_distances = []

        history_length = int(round(self.f * processing_config.history_length_s)) + 1
        self.detection_history = np.zeros(history_length) * float("nan")
        self.detection_history_t = np.linspace(-(history_length - 1) / self.f, 0, history_length)

    def process(self, data, data_info):
        sweep = data

        valid_leak_setup = (
            self.leak_sample_index >= 0
            and self.leak_sample_index < self.leak_end_index
            and self.leak_sample_index < len(sweep)
        )
        if valid_leak_setup:
            leak_amplitude = sweep[self.leak_sample_index]
            a_leak = max(leak_amplitude - ENVELOPE_BACKGROUND_LEVEL, 0)
            leak_step = a_leak / (self.leak_end_index - self.leak_sample_index)
            leak_start = self.leak_end_index * leak_step + ENVELOPE_BACKGROUND_LEVEL

            bg_near = np.linspace(leak_start, ENVELOPE_BACKGROUND_LEVEL, self.leak_end_index + 1)
            bg_far_len = len(sweep) - (self.leak_end_index + 1)
            if bg_far_len > 0:
                bg_far = np.ones(bg_far_len) * ENVELOPE_BACKGROUND_LEVEL
                background = np.append(bg_near, bg_far)
            else:
                background = bg_near[: len(sweep)]
        else:
            leak_amplitude = float("nan")
            background = np.ones(len(sweep)) * ENVELOPE_BACKGROUND_LEVEL

        leak_estimate = np.array([leak_amplitude, ENVELOPE_BACKGROUND_LEVEL])
        samples_above_bg = np.fmax(sweep - background, 0)
        weight = (
            np.fmin(samples_above_bg / ENVELOPE_BACKGROUND_LEVEL, 1)
            * samples_above_bg
            * self.norm_depths
        )

        weight_sum = np.sum(weight)

        sweep_weight = weight_sum / len(weight)
        sweep_distance = np.sum(weight * self.depths) / weight_sum

        # Pops the oldest item in the detector queue if the queue is full
        if len(self.queued_weights) == self.detector_queue_target_length:
            self.queued_weights = self.queued_weights[1:]
            self.queued_distances = self.queued_distances[1:]

        self.queued_weights.append(sweep_weight)
        self.queued_distances.append(sweep_distance)

        weight_min = min(self.queued_weights)
        weight_max = max(self.queued_weights)
        distance_min = min(self.queued_distances)
        distance_max = max(self.queued_distances)

        # The final criterion evaluation for parking detection
        detection = (
            len(self.queued_weights) == self.detector_queue_target_length
            and weight_min >= self.weight_threshold
            and weight_max / weight_min <= self.weight_ratio_limit
            and distance_max - distance_min <= self.distance_difference_limit
        )

        self.detection_history = np.roll(self.detection_history, -1)
        self.detection_history[-1] = detection

        # Calculates limits_center used to visualize the detection criterion
        limits_center = (np.sqrt(weight_min * weight_max), (distance_min + distance_max) / 2)

        out_data = {
            "sweep": sweep,
            "leak_estimate": leak_estimate,
            "leak_estimate_depths": self.leak_estimate_depths,
            "background": background,
            "weight": weight,
            "queued_weights": np.array(self.queued_weights),
            "queued_distances": np.array(self.queued_distances),
            "limits_center": limits_center,
            "detection_history": self.detection_history,
            "detection_history_t": self.detection_history_t,
        }

        return out_data


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 3

    depth_leak_sample = et.configbase.FloatParameter(
        label="Leak sample position",
        default_value=0.15,
        limits=(0.05, 0.25),
        unit="m",
        logscale=False,
        decimals=3,
        updateable=True,
        order=0,
        help="Distance from the sensor for the leak sample position",
    )

    depth_leak_end = et.configbase.FloatParameter(
        label="Leak end position",
        default_value=0.30,
        limits=(0.10, 0.50),
        unit="m",
        logscale=False,
        decimals=3,
        updateable=True,
        order=1,
        help="Worst case distance from the sensor for the end of leak reflections",
    )

    detector_queue_target_length = et.configbase.IntParameter(
        label="Detector queue length",
        default_value=3,
        limits=(1, 10),
        updateable=True,
        order=3,
        help=(
            "Car detection criterion parameter: "
            "The number of sweep value pairs (weight, distance) in the detector queue"
        ),
    )

    weight_threshold = et.configbase.FloatParameter(
        label="Weight threshold",
        default_value=5,
        limits=(0.5, 500),
        logscale=True,
        decimals=1,
        updateable=True,
        order=4,
        help=(
            "Car detection criterion parameter: "
            "Minimal value of the weights in the detector queue"
        ),
    )

    weight_ratio_limit = et.configbase.FloatParameter(
        label="Weight ratio limit",
        default_value=3,
        limits=(1, 10),
        logscale=True,
        decimals=2,
        updateable=True,
        order=5,
        help=(
            "Car detection criterion parameter: "
            "Maximal ratio between the maximal and the minimal weights in the detector queue"
        ),
    )

    distance_difference_limit = et.configbase.FloatParameter(
        label="Distance limit",
        default_value=0.2,
        limits=(0.01, 0.5),
        logscale=True,
        decimals=3,
        updateable=True,
        order=6,
        help=(
            "Car detection criterion parameter: "
            "Maximal difference between the maximal and minimal distances in the detector queue"
        ),
    )

    history_length_s = et.configbase.FloatParameter(
        label="History length",
        unit="s",
        default_value=300,
        limits=(1, 3600),
        logscale=True,
        decimals=0,
        updateable=True,
        order=100,
        help='The time interval that is shown in the "Detection history" plot',
    )

    def check(self):
        alerts = []

        if self.depth_leak_sample >= self.depth_leak_end:
            alerts.append(
                et.configbase.Error("depth_leak_sample", "Must be less than the leak end position")
            )

        return alerts

    def check_sensor_config(self, sensor_config):
        alerts = {
            "processing": [],
            "sensor": [],
        }
        if sensor_config.update_rate is None:
            alerts["sensor"].append(et.configbase.Error("update_rate", "Must be set"))

        if not sensor_config.noise_level_normalization:
            alerts["sensor"].append(
                et.configbase.Error("noise_level_normalization", "Must be set")
            )

        return alerts



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

from typing import List

from textual import events
from textual.app import App, Driver
from textual.geometry import Spacing
from textual.widgets import Button, ButtonPressed, Header, ScrollView

from ._dev_license import DevLicense


class DevLicenseTuiDialog(App):
    # The reason for this class variable is that the textual app is started
    # using the class method App#run, which in turn instantiate the subclass
    # DevLicenseTuiDialog. Because of this we don't (easily) get a handle to
    # the textual app instance.
    accepted = False

    def __init__(
        self,
        license: DevLicense,
        screen: bool = True,
        driver_class: type[Driver] | None = None,
        log: str = "",
        log_verbosity: int = 1,
        title: str = "",
    ) -> None:
        super().__init__(screen, driver_class, log, log_verbosity, title)

        self.license = license

        self.tab_order: List[Button] = []
        self.tab_index = -1

    @classmethod
    def get_accept(cls) -> bool:
        return cls.accepted

    async def on_load(self) -> None:
        await self.bind("q", "quit", "Quit")

    async def on_mount(self) -> None:
        self.title = self.license.get_header()
        header = Header(tall=False)
        await self.view.dock(header, edge="top")

        text_view = ScrollView(contents=self._get_license_text(), auto_width=False)

        user_consent_label = Button(
            label="To fetch the latest image file you must accept the "
            "terms and conditions stated by the license above!",
            style="bold white on rgb(50,57,50)",
        )

        user_consent_label.margin = Spacing(1, 0, 0, 0)

        self.accept_button = Button(
            label="Accept", name="accept", style="bold white on rgb(96,96,104)"
        )
        self.accept_button.margin = Spacing(1, 1, 1, 1)
        self.accept_button.border = "normal"

        self.reject_button = Button(
            label="Reject", name="reject", style="bold white on rgb(96,96,104)"
        )
        self.reject_button.margin = Spacing(1, 1, 1, 1)
        self.reject_button.border = "normal"

        self.tab_order.append(self.accept_button)
        self.tab_order.append(self.reject_button)

        await self.view.dock(text_view, size=20)
        await self.view.dock(user_consent_label, size=4)

        button_grid = await self.view.dock_grid(size=5)
        button_grid.add_column("col", repeat=2)
        button_grid.add_row("row", repeat=1)
        button_grid.set_align("stretch", "center")
        button_grid.set_gap(1, 1)

        button_grid.add_widget(self.accept_button)
        button_grid.add_widget(self.reject_button)

    def _get_license_text(self) -> str:
        license_header = [self.license.get_subheader(), "\n\n"]
        paragraphs = "\n\n".join(self.license.get_content())
        return "".join([*license_header, *paragraphs])

    async def handle_button_pressed(self, message: ButtonPressed) -> None:
        assert isinstance(message.sender, Button)
        await self.save_user_consent_and_exit(message.sender.name)

    async def save_user_consent_and_exit(self, button_name: str) -> None:
        if button_name == "accept":
            DevLicenseTuiDialog.accepted = True
            await self.close_messages()
        elif button_name == "reject":
            DevLicenseTuiDialog.accepted = False
            await self.close_messages()

    async def on_key(self, event: events.Key) -> None:
        if event.key == "ctrl+i":
            if self.tab_index == -1:
                self.tab_index = 0
            else:
                self.tab_index = (self.tab_index + 1) % len(self.tab_order)

            for index, button in enumerate(self.tab_order):
                if index != self.tab_index:
                    button.border = "normal"
                    button.render_styled()

            selected_button = self.tab_order[self.tab_index]
            selected_button.border = "bold"
            selected_button.render_styled()

        elif event.key == "enter":
            if self.tab_index != -1:
                selected_button = self.tab_order[self.tab_index]
                await self.save_user_consent_and_exit(selected_button.name)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import typing as t

import h5py

from acconeer.exptool._core.recording import h5_record
from acconeer.exptool._core.recording.h5_record.utils import PathOrH5File
from acconeer.exptool._core.recording.recorder import Recorder, RecorderAttachable
from acconeer.exptool.a121._core.entities import (
    Metadata,
    Result,
    SensorCalibration,
    ServerInfo,
    SessionConfig,
)

from .saver import H5Saver


T = t.TypeVar("T")


def get_h5py_str_dtype() -> t.Any:
    return h5py.special_dtype(vlen=str)


_H5PY_STR_DTYPE = get_h5py_str_dtype()


class H5Recorder(
    h5_record.H5Recorder[
        SessionConfig,  # Config type
        t.List[t.Dict[int, Metadata]],  # Metadata type
        t.List[t.Dict[int, Result]],  # Result type
        ServerInfo,  # Server info type
    ]
):
    """Recorder writing directly to an HDF5 file

    :param path_or_file:
        If a path-like object, by default (depending on ``mode`` below) an HDF5 file is created at
        that path. When stopping, the file is closed.

        If an ``h5py.File``, that file is used as-is. In this case, the file is not closed when
        stopping.
    :param attachable:
        A ``Client`` that the recorder will be attached to. When a recorder is attached to a
        ``Client``, all metadata and data frames will be recorded.
    :param mode:
        The file mode to use if a path-like object was given for ``path_or_file``. Default value is
        'x', meaning that we open for exclusive creation, failing if the file already exists.
    :param _chunk_size:
        If given, data will be written to file every ``_chunk_size`` samples.

        If not given, data will be written at least every 512:th sample, or at least once per
        second, whichever comes first.

        Setting a small chunk size (e.g. 1) may degrade performance for high sample rates.

        Internal parameter, subject to change.
    """

    def __init__(
        self,
        path_or_file: PathOrH5File,
        attachable: t.Optional[
            RecorderAttachable[
                Recorder[
                    SessionConfig,  # Config type
                    t.List[t.Dict[int, Metadata]],  # Metadata type
                    t.List[t.Dict[int, Result]],  # Result type
                    ServerInfo,  # Server info type
                ]
            ]
        ] = None,
        mode: str = "x",
        *,
        _chunk_size: t.Optional[int] = None,
        _lib_version: t.Optional[str] = None,
        _timestamp: t.Optional[str] = None,
        _uuid: t.Optional[str] = None,
    ) -> None:
        super().__init__(
            path_or_file,
            "a121",
            h5_record.ChunkedH5Saver(H5Saver(), _chunk_size=_chunk_size),
            attachable,
            mode,
            _lib_version=_lib_version,
            _timestamp=_timestamp,
            _uuid=_uuid,
        )

    def _start_session(
        self,
        *,
        config: SessionConfig,
        metadata: t.List[t.Dict[int, Metadata]],
        calibrations: t.Optional[t.Dict[int, SensorCalibration]] = None,
        calibrations_provided: t.Optional[t.Dict[int, bool]] = None,
        **kwargs: t.Any,
    ) -> None:
        super()._start_session(
            config=config,
            metadata=metadata,
            calibrations=calibrations,
            calibrations_provided=calibrations_provided,
            **kwargs,
        )



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import abc
import typing as t

import typing_extensions as te


class Message(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> te.Self:
        """Assumes that header and payload contains needed information and tries to parse.

        :param header: Message header
        :param payload: payload
        :raises Exception: whenever an error occurs (assumption doesn't hold)

        :returns: A freshly parsed Message
        """
        ...


class ParseError(Exception):
    pass



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved

from __future__ import annotations

import logging
import math
import time
from enum import Enum, auto
from typing import Any, Callable, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtWidgets import QLabel, QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121, opser
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo.distance._detector import DetailedStatus, Detector
from acconeer.exptool.a121.algo.distance._detector_plugin import ViewPlugin as DistanceViewPlugin
from acconeer.exptool.a121.algo.tank_level._configs import (
    get_large_config,
    get_medium_config,
    get_small_config,
)
from acconeer.exptool.a121.algo.tank_level._processor import ProcessorLevelStatus
from acconeer.exptool.a121.algo.tank_level._ref_app import (
    RefApp,
    RefAppConfig,
    RefAppContext,
    RefAppResult,
    _load_algo_data,
    ref_app_context_timeline,
)
from acconeer.exptool.app.new import (
    AppModel,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    HandledException,
    Message,
    PgPlotPlugin,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    PluginState,
    PluginStateMessage,
    backend,
    icons,
    is_task,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components import (
    AttrsConfigEditor,
    PidgetFactoryMapping,
    PidgetGroupFactoryMapping,
    PresentationType,
    pidgets,
)
from acconeer.exptool.app.new.ui.components.pidgets.hooks import (
    disable_if,
    parameter_is,
)


log = logging.getLogger(__name__)


NO_DETECTION_TIMEOUT = 50
TIME_HISTORY_S = 30
UPDATE_RATE_EXP_FILTER = 0.1


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: RefAppConfig = attrs.field(factory=RefAppConfig)
    context: RefAppContext = attrs.field(factory=RefAppContext)


class PluginPresetId(Enum):
    SMALL = auto()
    MEDIUM = auto()
    LARGE = auto()


@attrs.mutable(kw_only=True)
class TankLevelPreset:
    config: RefAppConfig = attrs.field()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    config: RefAppConfig
    num_curves: int
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, TankLevelPreset] = {
        PluginPresetId.SMALL.value: TankLevelPreset(
            config=get_small_config(),
        ),
        PluginPresetId.MEDIUM.value: TankLevelPreset(
            config=get_medium_config(),
        ),
        PluginPresetId.LARGE.value: TankLevelPreset(
            config=get_large_config(),
        ),
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key, use_app_client=False)

        self._recorder = None
        self._ref_app_instance: Optional[RefApp] = None
        self._log = BackendLogger.getLogger(__name__)
        self._last_get_next_time_s: Optional[float] = None
        self._actual_update_rate_hz = 0.0
        self._frame_count = 0

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = RefAppConfig.from_json(file["config"][()])
        self.shared_state.context = ref_app_context_timeline.migrate(file["context"])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState(config=get_small_config())
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: RefAppConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config.config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())
        context_group = file.create_group("context")
        opser.serialize(self.shared_state.context, context_group)

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        sensor_id, config, context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.context = context
        self.shared_state.sensor_id = sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client

        self._ref_app_instance = RefApp(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            config=self.shared_state.config,
            context=self.shared_state.context,
        )

        self._ref_app_instance.start(recorder)

        self.callback(
            SetupMessage(
                config=self.shared_state.config,
                num_curves=len(self._ref_app_instance._detector.processor_specs),
            )
        )

    def end_session(self) -> None:
        if self._ref_app_instance is None:
            raise RuntimeError

        if self._recorder is not None:
            self._recorder.close()

        self._ref_app_instance.stop()

        self._clear_rate_stats()

    def get_next(self) -> None:
        assert self.client is not None
        if self._ref_app_instance is None:
            raise RuntimeError

        sleep_time = self._sleep_until_next_frame()

        self._actual_update_rate_hz = self._iteratively_estimate_update_rate(
            last_get_next_time_s=self._last_get_next_time_s,
            current_estimated_update_rate=self._actual_update_rate_hz,
        )

        self._last_get_next_time_s = time.perf_counter()
        result = self._ref_app_instance.get_next()

        # Report frame count and update-rate to GUI.
        # usually done in ApplicationClient but because of erratic behavior
        # when distance detector is re-calibrated, it has to be done by the
        # ref app plugin instead.
        self._update_rate_stats(sleep_time=sleep_time, result=result)

    @is_task
    def calibrate_detector(self) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError

        if not self.client.connected:
            raise RuntimeError

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

        try:
            self._ref_app_instance = RefApp(
                client=self.client,
                sensor_id=self.shared_state.sensor_id,
                config=self.shared_state.config,
                context=None,
            )
            self._ref_app_instance.calibrate()
        except Exception as exc:
            msg = "Failed to calibrate detector"
            raise HandledException(msg) from exc
        finally:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        self.shared_state.context = self._ref_app_instance._detector.context
        self.broadcast()

    def _sleep_until_next_frame(self) -> Optional[float]:
        if self._ref_app_instance is None:
            raise RuntimeError

        if self._ref_app_instance.config.update_rate is None:
            return None

        if self._last_get_next_time_s is None:
            return None

        sleep_time = (
            self._last_get_next_time_s
            + 1 / self._ref_app_instance.config.update_rate
            - time.perf_counter()
        )

        if sleep_time > 0:
            time.sleep(sleep_time)

        return sleep_time

    @staticmethod
    def _iteratively_estimate_update_rate(
        last_get_next_time_s: Optional[float], current_estimated_update_rate: float
    ) -> float:
        if last_get_next_time_s is None:
            return 0.0

        actual_update_rate_hz = UPDATE_RATE_EXP_FILTER * current_estimated_update_rate + (
            1 - UPDATE_RATE_EXP_FILTER
        ) / (time.perf_counter() - last_get_next_time_s)

        return actual_update_rate_hz

    def _update_rate_stats(self, sleep_time: Optional[float], result: RefAppResult) -> None:
        if sleep_time is not None:
            if sleep_time > 0:  # had time to sleep -> No rate warning
                stats = backend._RateStats(self._actual_update_rate_hz, False, math.nan, False)
                self.callback(GeneralMessage(name="rate_stats", data=stats))
            else:  # no time to sleep -> rate warning
                stats = backend._RateStats(self._actual_update_rate_hz, True, math.nan, False)
                self.callback(GeneralMessage(name="rate_stats", data=stats))
        else:  # Run as fast as possible -> no rate warning
            stats = backend._RateStats(self._actual_update_rate_hz, False, math.nan, False)
            self.callback(GeneralMessage(name="rate_stats", data=stats))

        for value in result.extra_result.detector_result.values():
            self._frame_count += len(
                value.service_extended_result
            )  # count frames in extended result
        self.callback(backend.PlotMessage(result=result))
        self.callback(GeneralMessage(name="frame_count", data=self._frame_count))

    def _clear_rate_stats(self) -> None:
        self._last_get_next_time_s = None
        self._actual_update_rate_hz = 0.0
        self._frame_count = 0
        self.callback(GeneralMessage(name="rate_stats", data=None))
        self.callback(GeneralMessage(name="frame_count", data=None))


class PlotPlugin(PgPlotPlugin):
    STATUS_MSG_MAP = {
        ProcessorLevelStatus.IN_RANGE: "In range",
        ProcessorLevelStatus.NO_DETECTION: "Not available",
        ProcessorLevelStatus.OVERFLOW: "Warning: Overflow",
        ProcessorLevelStatus.OUT_OF_RANGE: "Out of range",
    }

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)

        self.counter = 0
        self.bar_loc = 0

        self._plot_job: Optional[RefAppResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(message.config, message.num_curves)
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        config: RefAppConfig,
        num_curves: int,
    ) -> None:
        self.plot_layout.clear()

        self.num_curves = num_curves
        self.start_m = config.start_m
        self.end_m = config.end_m

        win = self.plot_layout

        # Sweep plot
        self.sweep_plot = win.addPlot(row=1, col=0, colspan=3)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance (m)")
        self.sweep_plot.addItem(pg.PlotDataItem())

        self.vertical_line_start = pg.InfiniteLine(
            pen=et.utils.pg_pen_cycler(2),
            label="Tank start",
            labelOpts={
                "position": 0.5,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.sweep_plot.addItem(self.vertical_line_start)
        self.vertical_line_end = pg.InfiniteLine(
            pen=et.utils.pg_pen_cycler(2),
            label="Tank end",
            labelOpts={
                "position": 0.5,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.sweep_plot.addItem(self.vertical_line_end)

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.sweep_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        pen = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.threshold_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        sweep_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        sweep_plot_legend.setParentItem(self.sweep_plot)
        sweep_plot_legend.addItem(self.sweep_curves[0], "Sweep")
        sweep_plot_legend.addItem(self.threshold_curves[0], "Threshold")

        # Level history plot
        self.level_history_plot = win.addPlot(row=0, col=1, colspan=2)
        self.level_history_plot.setMenuEnabled(False)
        self.level_history_plot.showGrid(x=True, y=True)
        self.level_history_plot.addLegend()
        self.level_history_plot.setLabel("left", "Estimated level (cm)")
        self.level_history_plot.setLabel("bottom", "Time (s)")
        self.level_history_plot.addItem(pg.PlotDataItem())

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.level_history_curve = self.level_history_plot.plot(**feat_kw)

        self.sweep_smooth_max = et.utils.SmoothMax()
        self.distance_hist_smooth_lim = et.utils.SmoothLimits()

        # Rect plot
        self.num_rects = 15
        self.rect_plot = pg.PlotItem()
        self.rect_plot.setAspectLocked()
        self.rect_plot.hideAxis("left")
        self.rect_plot.hideAxis("bottom")
        self.rects = []

        pen = pg.mkPen(None)
        rect_width = self.num_rects / 2.0
        for r in np.arange(self.num_rects) + 1:
            rect = pg.QtWidgets.QGraphicsRectItem(0, r, rect_width, 1)
            rect.setPen(pen)
            self.rect_plot.addItem(rect)
            self.rects.append(rect)

        self.plot_layout.addItem(self.rect_plot, row=0, col=0)

        # text items
        self.level_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:12pt;">'
            "{}</span></div>"
        )

        self.level_text_item = pg.TextItem(
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.rect_plot.addItem(self.level_text_item)
        self.level_text_item.hide()

    def draw_plot_job(
        self,
        *,
        result: RefAppResult,
    ) -> None:
        # Get the first element as the plugin only supports single sensor operation.
        (detector_result,) = list(result.extra_result.detector_result.values())
        assert detector_result.distances is not None

        time_and_level_dict = (
            result.extra_result.processor_extra_result.level_and_time_for_plotting
        )

        # clear sweep plots
        for idx, _ in enumerate(self.sweep_curves):
            self.sweep_curves[idx].clear()
            self.threshold_curves[idx].clear()
        # update sweep plot
        max_val_in_plot = 0
        for idx, processor_result in enumerate(detector_result.processor_results):
            assert processor_result.extra_result.used_threshold is not None
            assert processor_result.extra_result.distances_m is not None
            assert processor_result.extra_result.abs_sweep is not None

            self.sweep_curves[idx].setData(
                processor_result.extra_result.distances_m, processor_result.extra_result.abs_sweep
            )

            self.threshold_curves[idx].setData(
                processor_result.extra_result.distances_m,
                processor_result.extra_result.used_threshold,
            )

            max_val_in_subsweep = max(
                max(processor_result.extra_result.used_threshold),
                max(processor_result.extra_result.abs_sweep),
            )

            max_val_in_plot = max(max_val_in_plot, max_val_in_subsweep)

        self.sweep_plot.setYRange(0, self.sweep_smooth_max.update(max_val_in_plot))
        self.vertical_line_start.setValue(self.start_m)
        self.vertical_line_end.setValue(self.end_m)
        self.vertical_line_start.show()
        self.vertical_line_end.show()

        # update level history plot
        if any(~np.isnan(time_and_level_dict["level"])):
            self.level_history_curve.setData(
                time_and_level_dict["time"], time_and_level_dict["level"] * 100
            )
            self.level_history_plot.setXRange(-TIME_HISTORY_S + 1, 0)
            self.level_history_plot.setYRange(0, (self.end_m - self.start_m + 0.01) * 100)

        # update level plot
        if (
            result.level is not None
            and result.peak_detected is not None
            and result.peak_status is not None
        ):
            current_level = result.level
            peak_detected = result.peak_detected
            peak_status = result.peak_status

            # Show the percentage level plot if the plot width is greater than 600 pixels,
            # otherwise display the level as text.
            if self.plot_layout.width() >= 600:
                level_text = self.STATUS_MSG_MAP[peak_status]
                if peak_status == ProcessorLevelStatus.OVERFLOW:
                    for rect in self.rects:
                        rect.setBrush(et.utils.pg_brush_cycler(0))
                elif peak_detected and (peak_status == ProcessorLevelStatus.IN_RANGE):
                    self.bar_loc = round(
                        current_level / (self.end_m - self.start_m) * self.num_rects
                    )
                    for rect in self.rects[: self.bar_loc]:
                        rect.setBrush(et.utils.pg_brush_cycler(0))

                    for rect in self.rects[self.bar_loc :]:
                        rect.setBrush(et.utils.pg_brush_cycler(1))

                    level_text = "Level: {:.1f} cm, {:.0f} %".format(
                        current_level * 100,
                        current_level / (self.end_m - self.start_m) * 100,
                    )
                    self.counter = 0
                elif peak_detected and (peak_status == ProcessorLevelStatus.OUT_OF_RANGE):
                    for rect in self.rects:
                        rect.setBrush(et.utils.pg_brush_cycler(1))
                elif (not peak_detected) and (self.counter <= NO_DETECTION_TIMEOUT):
                    for rect in self.rects[: self.bar_loc]:
                        rect.setBrush(et.utils.pg_brush_cycler(0))

                    for rect in self.rects[self.bar_loc :]:
                        rect.setBrush(et.utils.pg_brush_cycler(1))

                    self.counter += 1
                else:
                    for rect in self.rects:
                        rect.setBrush(et.utils.pg_brush_cycler(1))
                        self.bar_loc = 0
                for rect in self.rects:
                    rect.setVisible(True)

                level_html = self.level_html_format.format(level_text)
                self.level_text_item.setHtml(level_html)
                self.level_text_item.setPos(self.num_rects / 4.0, self.num_rects + 4.0)
                self.level_text_item.show()
            else:
                level_text = self.STATUS_MSG_MAP[peak_status]
                if peak_detected and (peak_status == ProcessorLevelStatus.IN_RANGE):
                    level_text = "Level: {:.1f} cm, {:.0f} %".format(
                        current_level * 100,
                        current_level / (self.end_m - self.start_m) * 100,
                    )
                    self.counter = 0
                level_html = self.level_html_format.format(level_text)
                self.level_text_item.setHtml(level_html)
                self.level_text_item.setPos(self.num_rects / 4.0, self.num_rects + 4.0)
                self.level_text_item.show()
                for rect in self.rects:
                    rect.setVisible(False)


class ViewPlugin(A121ViewPluginBase):
    TEXT_MSG_MAP = {
        DetailedStatus.OK: "Ready to start.",
        DetailedStatus.CONTEXT_MISSING: "Run detector calibration.",
        DetailedStatus.CALIBRATION_MISSING: "Run detector calibration.",
        DetailedStatus.CONFIG_MISMATCH: (
            "Current configuration does not match the configuration "
            + "used during detector calibration. Run detector calibration."
        ),
    }

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self.app_model = app_model
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.calibrate_detector_button = QPushButton(icons.CALIBRATE(), "Calibrate detector")
        self.calibrate_detector_button.clicked.connect(self._on_calibrate_detector)

        self.defaults_button = QPushButton(icons.RESTORE(), "Reset settings and calibrations")
        self.defaults_button.clicked.connect(self._send_defaults_request)

        self.message_box = QLabel(self.sticky_widget)
        self.message_box.setWordWrap(True)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.calibrate_detector_button, 1, 0, 1, -1)
        button_group.layout().addWidget(self.defaults_button, 3, 0, 1, -1)
        button_group.layout().addWidget(self.message_box, 4, 0, 1, -1)

        sticky_layout.addWidget(button_group)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.tank_level_config_editor = AttrsConfigEditor(
            title="Tank level indicator parameters",
            factory_mapping=self._get_processor_pidget_mapping(),
            config_type=RefAppConfig,
            extra_presenter=_set_config_presenter,
            parent=self.scrolly_widget,
        )
        self.tank_level_config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.tank_level_config_editor)

        self.config_editor = AttrsConfigEditor(
            title="Detector parameters",
            factory_mapping=self._get_detector_pidget_mapping(),
            config_type=RefAppConfig,
            save_load_buttons=False,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_processor_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        partial_range_params = {
            "partial_tracking_range_m": pidgets.FloatPidgetFactory(
                name_label_text="Partial tracking range:",
                name_label_tooltip=get_attribute_docstring(
                    RefAppConfig, "partial_tracking_range_m"
                ),
                suffix=" m",
                decimals=3,
                limits=(0, 23),
            ),
        }
        return {
            pidgets.FlatPidgetGroup(): {
                "median_filter_length": pidgets.IntPidgetFactory(
                    name_label_text="Median filter length:",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "median_filter_length"
                    ),
                    limits=(1, 10),
                ),
                "num_medians_to_average": pidgets.IntPidgetFactory(
                    name_label_text="Num measurements averaged:",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "num_medians_to_average"
                    ),
                    limits=(1, 10),
                ),
                "start_m": pidgets.FloatPidgetFactory(
                    name_label_text="Tank start:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "start_m"),
                    suffix=" m",
                    decimals=3,
                    limits=(0.03, 23),
                ),
                "end_m": pidgets.FloatPidgetFactory(
                    name_label_text="Tank end:",
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "end_m"),
                    suffix=" m",
                    decimals=3,
                    limits=(0.05, 23),
                ),
                "level_tracking_active": pidgets.CheckboxPidgetFactory(
                    name_label_text="Level tracking active",
                    name_label_tooltip=get_attribute_docstring(
                        RefAppConfig, "level_tracking_active"
                    ),
                ),
            },
            pidgets.FlatPidgetGroup(
                hooks=disable_if(parameter_is("level_tracking_active", False)),
            ): partial_range_params,
            pidgets.FlatPidgetGroup(): {
                "update_rate": pidgets.OptionalFloatPidgetFactory(
                    name_label_tooltip=get_attribute_docstring(RefAppConfig, "update_rate"),
                    name_label_text="Update rate:",
                    checkbox_label_text="Limit",
                    suffix=" Hz",
                    limits=(0.5, None),
                    init_set_value=1,
                ),
            },
        }

    @classmethod
    def _get_detector_pidget_mapping(cls) -> PidgetFactoryMapping:
        COMMON_PIDGETS = {
            "max_profile",
            "reflector_shape",
            "peaksorting_method",
            "threshold_method",
            "fixed_threshold_value",
            "fixed_strength_threshold_value",
            "num_frames_in_recorded_threshold",
            "threshold_sensitivity",
            "signal_quality",
            "close_range_leakage_cancellation",
        }

        return {
            "max_step_length": pidgets.OptionalIntPidgetFactory(
                name_label_text="Max step length:",
                name_label_tooltip=get_attribute_docstring(RefAppConfig, "max_step_length"),
                checkbox_label_text="Set",
                limits=(1, None),
                init_set_value=1,
            ),
            **{
                aspect: factory
                for aspect, factory in DistanceViewPlugin.get_pidget_mapping().items()
                if aspect in COMMON_PIDGETS
            },
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.tank_level_config_editor.set_data(None)
            self.message_box.setText("")
        else:
            self.sensor_id_pidget.set_data(state.sensor_id)
            self.config_editor.set_data(state.config)
            self.tank_level_config_editor.set_data(state.config)

            detector_status = Detector.get_detector_status(
                state.config.to_detector_config(), state.context, [state.sensor_id]
            )

            self.message_box.setText(self.TEXT_MSG_MAP[detector_status.detector_state])

            results = state.config._collect_validation_results()

            not_handled = self.tank_level_config_editor.handle_validation_results(results)
            not_handled = self.config_editor.handle_validation_results(not_handled)

            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[
                self.defaults_button,
                self.config_editor,
                self.tank_level_config_editor,
                self.sensor_id_pidget,
            ],
        )

        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

        state = app_model.backend_plugin_state

        if state is None:
            detector_ready = False
            config_valid = False
        else:
            detector_ready = Detector.get_detector_status(
                state.config.to_detector_config(), state.context, [state.sensor_id]
            ).ready_to_start

            config_valid = (
                self._config_valid(state.config)
                and self.config_editor.is_ready
                and self.tank_level_config_editor.is_ready
            )

        self.calibrate_detector_button.setEnabled(
            visual_policies.start_button_enabled(app_model, extra_condition=config_valid)
        )
        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=detector_ready and config_valid
            )
        )

    def _config_valid(self, config: Optional[RefAppConfig]) -> bool:
        if config is None:
            return False

        try:
            config.validate()
        except a121.ValidationError:
            return False
        else:
            return True

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)

    def _on_config_update(self, config: RefAppConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    # TODO: move to detector base (?)
    def _send_start_request(self) -> None:
        BackendPlugin.start_session.rpc(
            self.app_model.put_task, with_recorder=self.app_model.recording_enabled
        )

    # TODO: move to detector base (?)
    def _send_stop_request(self) -> None:
        BackendPlugin.stop_session.rpc(self.app_model.put_task)

    def _on_calibrate_detector(self) -> None:
        BackendPlugin.calibrate_detector.rpc(self.app_model.put_task)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)


def _set_config_presenter(instance: Any, presentation_type: PresentationType) -> Optional[str]:
    if isinstance(instance, RefAppConfig) and presentation_type is PresentationType.C_SET_CONFIG:
        config: RefAppConfig = instance
        distance_config = config.to_detector_config()

        return f"""
static void set_config(acc_ref_app_tank_level_config_t *config, tank_level_preset_config_t preset)
{{
    // This snippet is generated to be compatible with RSS A121 v1.0.0
    // If there is a version missmatch the snippet might need some modification

    (void)preset;

    config->tank_range_start_m     = {config.start_m:.3f}f;
    config->tank_range_end_m       = {config.end_m:.3f}f;
    config->median_filter_length   = {config.median_filter_length}U;
    config->num_medians_to_average = {config.num_medians_to_average}U;

    acc_detector_distance_config_start_set(config->distance_config, {distance_config.start_m:.3f}f);
    acc_detector_distance_config_end_set(config->distance_config, {distance_config.end_m:.3f}f);
    acc_detector_distance_config_max_step_length_set(config->distance_config, {distance_config.max_step_length or 0}U);
    acc_detector_distance_config_max_profile_set(config->distance_config, ACC_CONFIG_{distance_config.max_profile.name});
    acc_detector_distance_config_num_frames_recorded_threshold_set(config->distance_config, {distance_config.num_frames_in_recorded_threshold}U);
    acc_detector_distance_config_peak_sorting_set(config->distance_config, ACC_DETECTOR_DISTANCE_PEAK_SORTING_{distance_config.peaksorting_method.name});
    acc_detector_distance_config_reflector_shape_set(config->distance_config, ACC_DETECTOR_DISTANCE_REFLECTOR_SHAPE_{distance_config.reflector_shape.name});
    acc_detector_distance_config_threshold_sensitivity_set(config->distance_config, {distance_config.threshold_sensitivity:.3f}f);
    acc_detector_distance_config_signal_quality_set(config->distance_config, {distance_config.signal_quality:.3f}f);
    acc_detector_distance_config_close_range_leakage_cancellation_set(config->distance_config, {str(distance_config.close_range_leakage_cancellation).lower()});
}}
"""

    return None


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


TANK_LEVEL_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="tank_level",
    title="Tank level",
    docs_link="https://docs.acconeer.com/en/latest/ref_apps/a121/tank_level.html",
    description="Measure liquid levels in tanks",
    family=PluginFamily.REF_APP,
    presets=[
        PluginPresetBase(name="Small", preset_id=PluginPresetId.SMALL),
        PluginPresetBase(name="Medium", preset_id=PluginPresetId.MEDIUM),
        PluginPresetBase(name="Large", preset_id=PluginPresetId.LARGE),
    ],
    default_preset_id=PluginPresetId.SMALL,
)



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import itertools
import typing as t

from PySide6.QtGui import QTextOption
from PySide6.QtWidgets import QTextEdit

from acconeer.exptool.a121.model import memory
from acconeer.exptool.app.new.ui.resource_tab.event_system import (
    ChangeIdEvent,
    EventBroker,
    IdentifiedServiceUninstalledEvent,
)

from .distance_config_input import DistanceConfigEvent
from .presence_config_input import PresenceConfigEvent
from .session_config_input import SessionConfigEvent


class MemoryBreakdownOutput(QTextEdit):
    INTERESTS: t.ClassVar[set[type]] = {
        DistanceConfigEvent,
        SessionConfigEvent,
        PresenceConfigEvent,
        IdentifiedServiceUninstalledEvent,
        ChangeIdEvent,
    }
    description: t.ClassVar[str] = "Tabulates heap memory consumption of configurations"
    window_title = "Memory breakdown"

    def __init__(self, broker: EventBroker) -> None:
        super().__init__()

        self._memory_numbers: dict[str, tuple[int, int]] = {}

        self.setPlaceholderText("Nothing to show")
        self.setReadOnly(True)
        self.setWordWrapMode(QTextOption.WrapMode.NoWrap)
        self.setFontFamily("monospace")

        broker.install_service(self)
        self.uninstall_function = lambda: broker.uninstall_service(self)
        broker.brief_service(self)

    def handle_event(self, event: t.Any) -> None:
        if isinstance(event, SessionConfigEvent):
            self._handle_session_config_event(event)
        elif isinstance(event, DistanceConfigEvent):
            self._handle_distance_config_event(event)
        elif isinstance(event, PresenceConfigEvent):
            self._handle_presence_config_event(event)
        elif isinstance(event, IdentifiedServiceUninstalledEvent):
            self._handle_identified_service_uninstalled_event(event)
        elif isinstance(event, ChangeIdEvent):
            self._handle_change_id_event(event)
        else:
            raise NotImplementedError

    def _handle_session_config_event(self, event: SessionConfigEvent) -> None:
        self._memory_numbers[event.service_id] = (
            memory.session_rss_heap_memory(event.session_config),
            memory.session_external_heap_memory(event.session_config),
        )

        self._show_memory_numbers()

    def _handle_distance_config_event(self, event: DistanceConfigEvent) -> None:
        self._memory_numbers[event.service_id] = (
            memory.distance_rss_heap_memory(event.config),
            memory.distance_external_heap_memory(event.config),
        )

        self._show_memory_numbers()

    def _handle_presence_config_event(self, event: PresenceConfigEvent) -> None:
        self._memory_numbers[event.service_id] = (
            memory.presence_rss_heap_memory(event.config),
            memory.presence_external_heap_memory(event.config),
        )

        self._show_memory_numbers()

    def _handle_identified_service_uninstalled_event(
        self, event: IdentifiedServiceUninstalledEvent
    ) -> None:
        self._memory_numbers.pop(event.id_)
        self._show_memory_numbers()

    def _handle_change_id_event(self, event: ChangeIdEvent) -> None:
        memory_entry = self._memory_numbers.pop(event.old_id, None)

        if memory_entry is not None:
            self._memory_numbers[event.new_id] = memory_entry

        self._show_memory_numbers()

    @staticmethod
    def _fmt_number(number: int) -> str:
        return str(number + (100 - number % 100)) + "B"

    def _show_memory_numbers(self) -> None:
        if not self._memory_numbers:
            self.setText("")
        else:
            longest_key = max(len(key) for key in self._memory_numbers)
            fmt = f"{{: <{longest_key}}} {{: <12}} {{}}"
            header = fmt.format("", "RSS heap", "External heap")

            self.setText(
                "\n".join(
                    itertools.chain(
                        [header, "=" * len(header)],
                        (
                            fmt.format(
                                service_id, self._fmt_number(rss), self._fmt_number(external)
                            )
                            for service_id, (rss, external) in self._memory_numbers.items()
                        ),
                    )
                )
            )



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from .communication import Client, ClientCreationError, ClientError
from .entities import (
    ClientInfo,
    MockInfo,
    SerialInfo,
    SocketInfo,
    USBInfo,
)
from .int_16_complex import (
    INT_16_COMPLEX,
    complex_array_to_int16_complex,
    int16_complex_array_to_complex,
)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

import contextlib
import multiprocessing as mp
import queue
import signal
import sys
from time import sleep, time

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation


class PlotProcess:
    def __init__(self, fig_updater, interval=10):
        self._queue = mp.Queue()
        self._exit_event = mp.Event()

        args = (
            self._queue,
            self._exit_event,
            fig_updater,
            interval / 1000.0,
        )

        self._process = mp.Process(target=plot_process_program, args=args, daemon=True)

    def start(self):
        self._process.start()

    def put_data(self, data):
        if self._process.exitcode is not None:
            self.close()
            raise PlotProccessDiedException

        try:
            self._queue.put(data)
        except BrokenPipeError:
            self.close()
            raise PlotProccessDiedException

    def close(self):
        if self._process.exitcode is None:
            self._exit_event.set()
            self._process.join(1)

        if self._process.exitcode is None:
            self._process.terminate()
            self._process.join(1)

        if self._process.exitcode is None:
            raise RuntimeError


def plot_process_program(q, exit_event, fig_updater, interval):
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    last_t = None
    artists = []
    fig = plt.figure()

    def anim_func(frame):
        nonlocal artists, last_t

        if exit_event.is_set():
            sys.exit()

        if last_t:
            now = time()
            sleep_t = (last_t + interval - 0.001) - now
            if sleep_t > 0.001:
                sleep(sleep_t)
        last_t = time()

        data = None
        while True:
            try:
                data = q.get(timeout=0.001)
            except queue.Empty:
                break

        if data is not None:
            if not artists:
                artists = fig_updater.first(data)
            else:
                fig_updater.update(data)

        return artists

    fig_updater.setup(fig)
    anim = FuncAnimation(fig, anim_func, interval=0, blit=True)  # noqa: F841

    with contextlib.suppress(AttributeError):
        plt.show()


class PlotProccessDiedException(Exception):
    pass


class FigureUpdater:
    def setup(self, fig):
        self.fig = fig

    def first(self, data):
        return []  # iterable of artists

    def update(self, data):
        pass


class ExampleFigureUpdater(FigureUpdater):
    def setup(self, fig):
        self.fig = fig
        self.ax = fig.subplots()
        self.ax.set_xlabel("t")
        self.ax.set_ylabel("sin(t)")

    def first(self, data):
        x, y = data
        self.line = self.ax.plot(x, y)[0]
        return [self.line]

    def update(self, data):
        _, y = data
        self.line.set_ydata(y)


# Example usage:
if __name__ == "__main__":
    import numpy as np

    fu = ExampleFigureUpdater()
    pp = PlotProcess(fu)
    pp.start()

    x = np.linspace(0, 10, 100)
    t0 = time()
    t = 0
    while t < 10:
        t = time() - t0
        y = np.sin(x + t)

        try:
            pp.put_data([x, y])
        except PlotProccessDiedException:
            exit()

        sleep(0.001)  # not necessary

    pp.close()



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from .client_info import (
    ClientInfo,
    ClientInfoCreationError,
    MockInfo,
    SerialInfo,
    SocketInfo,
    USBInfo,
)
from .validation_result import (
    Criticality,
    ValidationError,
    ValidationResult,
    ValidationWarning,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Any, Dict


class ExampleArgumentParser(ArgumentParser):
    def __init__(self) -> None:
        super().__init__()

        server_group = self.add_mutually_exclusive_group(required=False)
        server_group.add_argument(
            "--serial-port",
            dest="serial_port",
            metavar="port",
        )
        server_group.add_argument(
            "--ip-address",
            dest="ip_address",
            metavar="address",
        )
        server_group.add_argument(
            "--usb-device",
            dest="usb_device",
            metavar="serial_number",
            nargs="?",
            const=True,
        )
        server_group.add_argument(
            "--mock",
            dest="mock",
            default=None,
            action="store_true",
        )

        verbosity_group = self.add_mutually_exclusive_group(required=False)
        verbosity_group.add_argument(
            "-v",
            "--verbose",
            action="store_true",
        )
        verbosity_group.add_argument(
            "-vv",
            "--debug",
            action="store_true",
        )
        verbosity_group.add_argument(
            "-q",
            "--quiet",
            action="store_true",
        )


def get_client_args(namespace: Namespace) -> Dict[str, Any]:
    """
    Filters a passed Namespace to extract client-creation related args.

    :returns: dictionary with client-related keyword-arguments.
    """

    CLIENT_RELATED_KEYS = ["ip_address", "serial_port", "usb_device", "mock"]
    return {k: getattr(namespace, k) for k in CLIENT_RELATED_KEYS if hasattr(namespace, k)}



# Copyright (c) Acconeer AB, 2025
# All rights reserved

from __future__ import annotations

import warnings
from enum import Enum
from typing import Any, List, Optional, Tuple, Union

import attrs
import h5py
from attributes_doc import attributes_doc

from acconeer.exptool import a121, opser
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    AlgoConfigBase,
    AlgoParamEnum,
    Controller,
)
from acconeer.exptool.a121.algo._base import AlgoBase, AlgoBaseT
from acconeer.exptool.a121.algo._utils import estimate_frame_rate
from acconeer.exptool.a121.algo.distance._detector import Detector as DistanceDetector
from acconeer.exptool.a121.algo.distance._detector import DetectorConfig as DistanceConfig
from acconeer.exptool.a121.algo.distance._detector import DetectorContext as DistanceContext
from acconeer.exptool.a121.algo.distance._detector import DetectorResult as DistanceResult
from acconeer.exptool.a121.algo.distance._detector import ReflectorShape
from acconeer.exptool.a121.algo.distance._processors import (
    ProcessorResult as DistanceProcessorResult,
)
from acconeer.exptool.a121.algo.presence._detector import Detector as PresenceDetector
from acconeer.exptool.a121.algo.presence._detector import DetectorConfig as PresenceConfig
from acconeer.exptool.a121.algo.presence._detector import DetectorContext as PresenceContext
from acconeer.exptool.a121.algo.presence._detector import DetectorResult as PresenceResult
from acconeer.exptool.a121.algo.presence._processors import Processor as PresenceProcessor


PRESENCE_RUN_TIME_S = 5.0
MIN_PRESENCE_UPDATE_RATE = 1.0


def container_size_converter(container_size: ContainerSize) -> ContainerSize:
    return ContainerSize(container_size)


@attributes_doc
@attrs.mutable(kw_only=True)
class CargoPresenceConfig(AlgoConfigBase):
    burst_rate: float = attrs.field(default=0.1)
    """Sets the presence burst range in Hz.
    Each burst is 5 s long, hence 0.2 Hz is the highest possible burst rate."""

    update_rate: float = attrs.field(default=6.0)
    """Update rate in Hz."""

    signal_quality: float = attrs.field(default=20)
    """Signal quality metric (higher = better signal, lower = less power consumption)."""

    sweeps_per_frame: int = attrs.field(default=12)
    """Number of sweeps per frame."""

    inter_detection_threshold: float = attrs.field(default=4)
    """Detection threshold for the inter-frame presence detection."""

    intra_detection_threshold: float = attrs.field(default=2.75)
    """Detection threshold for the intra-frame presence detection."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if self.burst_rate > (1 / PRESENCE_RUN_TIME_S):
            validation_results.append(
                a121.ValidationError(
                    self,
                    "burst_rate",
                    f"Must not be > {1 / PRESENCE_RUN_TIME_S} Hz.",
                )
            )

        if self.update_rate < 1:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "update_rate",
                    f"Must be at least {MIN_PRESENCE_UPDATE_RATE} Hz.",
                )
            )

        if self.sweeps_per_frame <= PresenceProcessor.NOISE_ESTIMATION_DIFF_ORDER:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sweeps_per_frame",
                    f"Must be greater than {PresenceProcessor.NOISE_ESTIMATION_DIFF_ORDER}",
                )
            )

        return validation_results


@attributes_doc
@attrs.mutable(kw_only=True)
class UtilizationLevelConfig(AlgoConfigBase):
    update_rate: float = attrs.field(default=5)
    """Sets the detector update rate."""

    signal_quality: float = attrs.field(default=10)
    """Signal quality (dB).

    High quality equals higher HWAAS and better SNR but increases power consumption."""

    threshold_sensitivity: float = attrs.field(default=0)
    """Sensitivity of threshold.

    High sensitivity equals low detection threshold, low sensitivity equals high detection
    threshold."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        return []


class ContainerSize(AlgoParamEnum):
    #               length_m  Presence    Distance
    CONTAINER_10_FT = (3.0, (0.8, 2.7), (0.3, 3.5))
    CONTAINER_20_FT = (6.0, (0.8, 5.7), (0.3, 6.5))
    CONTAINER_40_FT = (12.0, (0.8, 7.0), (0.3, 12.5))

    @property
    def length_m(self) -> float:
        return self.value[0]

    @property
    def presence_range(self) -> tuple[float, float]:
        return self.value[1]

    @property
    def distance_range(self) -> tuple[float, float]:
        return self.value[2]


@attributes_doc
@attrs.mutable(kw_only=True)
class ExAppConfig(AlgoConfigBase):
    activate_presence: bool = attrs.field()
    """Activates presence detection."""

    cargo_presence_config: Optional[CargoPresenceConfig] = attrs.field(factory=CargoPresenceConfig)
    """Configuration of presence detector."""

    activate_utilization_level: bool = attrs.field()
    """Activates utilization level detection."""

    utilization_level_config: Optional[UtilizationLevelConfig] = attrs.field(
        factory=UtilizationLevelConfig
    )
    """Configuration for utilization level measurements."""

    container_size: ContainerSize = attrs.field(
        default=ContainerSize.CONTAINER_20_FT, converter=container_size_converter
    )
    """Container size. Sets the detection range for utilization level detection and for presence detection."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []
        if self.cargo_presence_config is not None:
            validation_results.extend(self.cargo_presence_config._collect_validation_results())

        if self.utilization_level_config is not None:
            validation_results.extend(self.utilization_level_config._collect_validation_results())

        if not self.activate_presence and not self.activate_utilization_level:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "activate_presence",
                    "One detector needs to be activated.",
                )
            )

        return validation_results

    @classmethod
    def from_dict(cls: type[AlgoBaseT], d: dict[str, Any]) -> AlgoBaseT:
        d["cargo_presence_config"] = CargoPresenceConfig.from_dict(d["cargo_presence_config"])
        d["utilization_level_config"] = UtilizationLevelConfig.from_dict(
            d["utilization_level_config"]
        )

        return cls(**d)


@attrs.mutable(kw_only=True)
class ExAppContext(AlgoBase):
    presence_context: Optional[PresenceContext] = attrs.field(default=None)
    distance_context: Optional[DistanceContext] = attrs.field(default=None)

    @classmethod
    def from_dict(cls: type[AlgoBaseT], d: dict[str, Any]) -> AlgoBaseT:
        if d["presence_context"] is not None:
            d["presence_context"] = PresenceContext.from_dict(d["presence_context"])
        if d["distance_context"] is not None:
            d["distance_context"] = DistanceContext.from_dict(d["distance_context"])

        return cls(**d)


@attrs.frozen(kw_only=True)
class ExAppResult:
    mode: _Mode = attrs.field()
    """Measurement mode, Presence or Distance."""

    presence_detected: Optional[bool] = attrs.field()
    """`True` if presence was detected, `False` otherwise."""

    inter_presence_score: Optional[float] = attrs.field()
    """A measure of the amount of slow motion detected."""

    intra_presence_score: Optional[float] = attrs.field()
    """A measure of the amount of fast motion detected."""

    distance: Optional[float] = attrs.field()
    """The measured distance by the distance detector."""

    level_m: Optional[float] = attrs.field()
    """The utilization level in m."""

    level_percent: Optional[float] = attrs.field()
    """The utilization level in percentage."""

    distance_processor_result: Optional[List[DistanceProcessorResult]] = attrs.field()
    """Distance processor result, used for plotting."""

    service_result: a121.Result = attrs.field()


class _Mode(Enum):
    PRESENCE = 0
    DISTANCE = 1


class ExApp(Controller[ExAppConfig, ExAppResult]):
    detector: Union[DistanceDetector, PresenceDetector]

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        ex_app_config: ExAppConfig,
        ex_app_context: Optional[ExAppContext] = None,
    ) -> None:
        super().__init__(client=client, config=ex_app_config)
        self.sensor_id = sensor_id
        if ex_app_context is None:
            self.ex_app_context = ExAppContext()
        else:
            self.ex_app_context = ex_app_context

        self.ex_app_config = ex_app_config
        self.presence_context = self.ex_app_context.presence_context
        self.distance_context = self.ex_app_context.distance_context

        self.container_size = ex_app_config.container_size
        self.activate_presence = ex_app_config.activate_presence
        self.cargo_presence_config = ex_app_config.cargo_presence_config
        self.activate_utilization_level = ex_app_config.activate_utilization_level
        self.utilization_level_config = ex_app_config.utilization_level_config

        if self.activate_presence and self.activate_utilization_level:
            self.dual_detectors = True
        elif not self.activate_presence and not self.activate_utilization_level:
            msg = "No detector activated"
            raise AssertionError(msg)
        else:
            self.dual_detectors = False

        self.started = False

    def start(
        self, recorder: Optional[a121.Recorder] = None, algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        if self.activate_presence:
            self.initialize_presence()

        if self.activate_utilization_level:
            self.initialize_distance()

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                algo_group = recorder.require_algo_group("cargo")
                _record_algo_data(
                    algo_group,
                    self.sensor_id,
                    self.ex_app_config,
                    self.ex_app_context,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

        if self.dual_detectors:
            self.start_dual_detectors(recorder, algo_group)
        elif self.activate_presence:
            self.detector = PresenceDetector(
                client=self.client,
                sensor_id=self.sensor_id,
                detector_config=self.presence_config,
                detector_context=self.presence_context,
            )
            self.detector.start(recorder=recorder, _algo_group=algo_group)
        else:
            self.detector = DistanceDetector(
                client=self.client,
                sensor_ids=[self.sensor_id],
                detector_config=self.distance_config,
                context=self.distance_context,
            )
            self.detector.start(recorder=recorder, _algo_group=algo_group)

        self.started = True

    def initialize_presence(self) -> None:
        assert self.cargo_presence_config is not None
        self.num_presence_frames_in_burst = int(
            round((PRESENCE_RUN_TIME_S * self.cargo_presence_config.update_rate))
        )
        self.num_presence_frames_in_cycle = (
            self.cargo_presence_config.update_rate / self.cargo_presence_config.burst_rate
        )
        self.presence_count = 0
        start_m, end_m = self.container_size.presence_range
        self.presence_config = self.get_presence_config(self.cargo_presence_config, start_m, end_m)

        sensor_config = PresenceDetector._get_sensor_config(self.presence_config)
        session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        if self.ex_app_context is None:
            self.presence_context = PresenceContext(
                estimated_frame_rate=estimate_frame_rate(self.client, session_config)
            )
            self.ex_app_context = ExAppContext(presence_context=self.presence_context)
        elif self.ex_app_context.presence_context is None:
            self.presence_context = PresenceContext(
                estimated_frame_rate=estimate_frame_rate(self.client, session_config)
            )
            self.ex_app_context.presence_context = self.presence_context
        else:
            self.presence_context = self.ex_app_context.presence_context

    def initialize_distance(self) -> None:
        assert self.utilization_level_config is not None
        start_m, end_m = self.container_size.distance_range
        self.distance_config = self.get_distance_config(
            self.utilization_level_config, start_m, end_m
        )

        if self.ex_app_context is None or self.ex_app_context.distance_context is None:
            self.detector = DistanceDetector(
                client=self.client,
                sensor_ids=[self.sensor_id],
                detector_config=self.distance_config,
                context=self.distance_context,
            )

            self.detector.calibrate_detector()
            if self.ex_app_context is None:
                self.ex_app_context = ExAppContext(distance_context=self.detector.context)
            else:
                self.ex_app_context.distance_context = self.detector.context

        self.distance_context = self.ex_app_context.distance_context

    def start_dual_detectors(
        self, recorder: Optional[a121.Recorder], algo_group: Optional[h5py.Group]
    ) -> None:
        assert self.cargo_presence_config is not None
        assert self.utilization_level_config is not None

        if self.utilization_level_config.update_rate > self.cargo_presence_config.burst_rate:
            self.burst_nbr_switch = (
                self.utilization_level_config.update_rate / self.cargo_presence_config.burst_rate
                - self.utilization_level_config.update_rate * PRESENCE_RUN_TIME_S
            )
            self.burst_nbr_switch = int(round(self.burst_nbr_switch))
            self.steering_mode = _Mode.DISTANCE

        else:
            self.burst_nbr_switch = (
                self.cargo_presence_config.burst_rate / self.utilization_level_config.update_rate
                - self.cargo_presence_config.burst_rate * PRESENCE_RUN_TIME_S
            )
            self.steering_mode = _Mode.PRESENCE

        self.detector = DistanceDetector(
            client=self.client,
            sensor_ids=[self.sensor_id],
            detector_config=self.distance_config,
            context=self.distance_context,
        )
        self.detector.start(recorder=recorder, _algo_group=algo_group)
        self.burst_count = 0
        self.presence_count = self.num_presence_frames_in_burst

    @classmethod
    def get_presence_config(
        cls, cargo_presence_config: CargoPresenceConfig, start_m: float, end_m: float
    ) -> PresenceConfig:
        return PresenceConfig(
            frame_rate=cargo_presence_config.update_rate,
            start_m=start_m,
            end_m=end_m,
            sweeps_per_frame=cargo_presence_config.sweeps_per_frame,
            automatic_subsweeps=True,
            signal_quality=cargo_presence_config.signal_quality,
            intra_detection_threshold=cargo_presence_config.intra_detection_threshold,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_detection_threshold=cargo_presence_config.inter_detection_threshold,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=cargo_presence_config.update_rate,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=None,
        )

    @classmethod
    def get_distance_config(
        cls, utilization_level_config: UtilizationLevelConfig, start_m: float, end_m: float
    ) -> DistanceConfig:
        return DistanceConfig(
            update_rate=utilization_level_config.update_rate,
            start_m=start_m,
            end_m=end_m,
            signal_quality=utilization_level_config.signal_quality,
            threshold_sensitivity=utilization_level_config.threshold_sensitivity,
            reflector_shape=ReflectorShape.PLANAR,
        )

    def get_next(self) -> ExAppResult:
        distance_result: Optional[dict[int, DistanceResult]] = None
        presence_result: Optional[PresenceResult] = None

        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        if isinstance(self.detector, PresenceDetector):
            presence_result = self.detector.get_next()
            current_mode = _Mode.PRESENCE
            self.presence_count += 1
        else:
            distance_result = self.detector.get_next()
            current_mode = _Mode.DISTANCE

        if self.dual_detectors:
            if self.presence_count < self.num_presence_frames_in_burst:
                skip_result = False
            else:
                skip_result = self.determine_swapping(current_mode)
        elif self.activate_presence:
            skip_result = False
            if (
                self.presence_count > self.num_presence_frames_in_burst
                and self.presence_count < self.num_presence_frames_in_cycle
            ):
                skip_result = True
            elif self.presence_count == self.num_presence_frames_in_cycle:
                if self.presence_count == self.num_presence_frames_in_burst:
                    skip_result = False
                else:
                    skip_result = True

                self.presence_count = 0
                self.detector.stop_detector()
                self.detector.start(recorder=None, _algo_group=None)

        if current_mode == _Mode.DISTANCE:
            assert distance_result is not None
            service_result = distance_result[self.sensor_id].service_extended_result[0][
                self.sensor_id
            ]
            presence_detected = None
            inter_presence_score = None
            intra_presence_score = None
            distances = distance_result[self.sensor_id].distances
            distance_processor_result = distance_result[self.sensor_id].processor_results
            if distances is not None and len(distances) != 0:
                distance = distances[0]
                level_m = self.ex_app_config.container_size.length_m - distance
                level_percent = level_m / self.ex_app_config.container_size.length_m * 100
            else:
                distance = None
                level_m = None
                level_percent = None
        else:
            assert presence_result is not None
            service_result = presence_result.service_result
            if skip_result:
                presence_detected = None
                inter_presence_score = None
                intra_presence_score = None
                distance = None
                level_m = None
                level_percent = None
                distance_processor_result = None
            else:
                presence_detected = presence_result.presence_detected
                inter_presence_score = presence_result.inter_presence_score
                intra_presence_score = presence_result.intra_presence_score
                distance = None
                level_m = None
                level_percent = None
                distance_processor_result = None

        return ExAppResult(
            mode=current_mode,
            presence_detected=presence_detected,
            inter_presence_score=inter_presence_score,
            intra_presence_score=intra_presence_score,
            distance=distance,
            level_m=level_m,
            level_percent=level_percent,
            distance_processor_result=distance_processor_result,
            service_result=service_result,
        )

    def determine_swapping(self, current_mode: _Mode) -> bool:
        skip_result = False
        if self.steering_mode == _Mode.DISTANCE:
            if current_mode == _Mode.PRESENCE:
                self.swap_config(current_mode)
            else:
                self.burst_count += 1
                if self.burst_count > self.burst_nbr_switch:
                    self.swap_config(current_mode)
                    self.presence_count = 0
        else:
            if current_mode == _Mode.DISTANCE:
                self.swap_config(current_mode)
                self.presence_count = 0
            else:
                if self.presence_count == self.num_presence_frames_in_cycle:
                    self.burst_count += 1
                    if self.burst_count > self.burst_nbr_switch:
                        self.swap_config(current_mode)
                    else:
                        self.presence_count = 0
                        self.detector.stop_detector()
                        self.detector.start(recorder=None, _algo_group=None)

                if self.presence_count > self.num_presence_frames_in_burst:
                    skip_result = True

        return skip_result

    def swap_config(self, current_mode: _Mode) -> None:
        self.detector.stop_detector()

        if current_mode == _Mode.DISTANCE:
            self.detector = PresenceDetector(
                client=self.client,
                sensor_id=self.sensor_id,
                detector_config=self.presence_config,
                detector_context=self.presence_context,
            )
        else:
            self.detector = DistanceDetector(
                client=self.client,
                sensor_ids=[self.sensor_id],
                detector_config=self.distance_config,
                context=self.distance_context,
            )

        self.detector.start(recorder=None, _algo_group=None)
        self.burst_count = 0

    def update_config(self, config: ExAppConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        recorder_result = self.detector.stop()

        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    ex_app_config: ExAppConfig,
    ex_app_context: ExAppContext,
) -> None:
    if "ex_app_sensor_id" in algo_group:
        algo_group.attrs["ex_app_sensor_id"] = sensor_id
        del algo_group["ex_app_config"]
        _create_h5_string_dataset(algo_group, "ex_app_config", ex_app_config.to_json())
        del algo_group["ex_app_context"]
        ex_app_context_group = algo_group.create_group("ex_app_context")
    else:
        algo_group.create_dataset(
            "ex_app_sensor_id",
            data=sensor_id,
            track_times=False,
        )

        _create_h5_string_dataset(algo_group, "ex_app_config", ex_app_config.to_json())
        ex_app_context_group = algo_group.create_group("ex_app_context")

    opser.serialize(ex_app_context, ex_app_context_group)


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, ExAppConfig, ExAppContext]:
    sensor_id = int(algo_group["ex_app_sensor_id"][()])
    config = ExAppConfig.from_json(algo_group["ex_app_config"][()])
    ex_app_context = opser.deserialize(algo_group["ex_app_context"], ExAppContext)
    return sensor_id, config, ex_app_context



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from acconeer.exptool.a121.algo.breathing._ref_app import RefAppConfig
from acconeer.exptool.a121.algo.presence import ProcessorConfig


def get_sitting_config() -> RefAppConfig:
    presence_config = ProcessorConfig()
    presence_config.intra_detection_threshold = 6.0

    ref_app_config = RefAppConfig()
    ref_app_config.end_m = 1.5
    ref_app_config.presence_config = presence_config

    return ref_app_config


def get_infant_config() -> RefAppConfig:
    presence_config = ProcessorConfig()
    presence_config.intra_detection_threshold = 4.0

    ref_app_config = RefAppConfig()
    ref_app_config.end_m = 1.0
    ref_app_config.presence_config = presence_config

    return ref_app_config



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from .launcher import run_launcher


if __name__ == "__main__":
    selected_generation = run_launcher()

    if selected_generation == "new":
        from .new import main

        main()
    elif selected_generation == "old":
        from .old import main

        main()



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import typing as t

import numpy as np
from numpy import typing as npt


INT_16_COMPLEX = np.dtype([("real", "int16"), ("imag", "int16")])


def int16_complex_array_to_complex(array: npt.NDArray[t.Any]) -> npt.NDArray[np.complex128]:
    """Converts an array with dtype = INT_16_COMPLEX
    (structured with parts "real" and "imag") into
    an array with plain complex dtype (non-structured).
    """
    real = array["real"].astype("float")
    imaginary = array["imag"].astype("float")
    return real + 1.0j * imaginary  # type: ignore[no-any-return]


def complex_array_to_int16_complex(array: npt.NDArray[np.complex128]) -> npt.NDArray[t.Any]:
    """Converts an array with plain complex dtype (non-structured)
    into an array with dtype = INT_16_COMPLEX
    (structured with parts "real" and "imag") using `numpy.round`.
    """
    struct_array = np.empty(array.shape, dtype=INT_16_COMPLEX)
    struct_array["real"] = np.round(array.real)
    struct_array["imag"] = np.round(array.imag)
    return struct_array



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from ._application_client import ApplicationClient
from ._backend import Backend, ClosedTask, GenBackend, MpBackend
from ._backend_logger import BackendLogger
from ._backend_plugin import BackendPlugin
from ._message import (
    BackendPluginStateMessage,
    ConnectionStateMessage,
    GeneralMessage,
    LogMessage,
    Message,
    PlotMessage,
    PluginStateMessage,
    RecipientLiteral,
    StatusFileAccessMessage,
    StatusMessage,
    TimingMessage,
)
from ._model import Model
from ._rate_calc import _RateCalculator, _RateStats
from ._tasks import Task, is_task



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import json
from typing import Any, Optional

from acconeer.exptool._core.communication import CommunicationProtocol, Message
from acconeer.exptool._core.communication.communication_protocol import messages
from acconeer.exptool.a121._core.entities import PRF, IdleState, SensorCalibration, SessionConfig
from acconeer.exptool.a121._core.utils import map_over_extended_structure

from .messages import EmptyResultMessage, ResultMessage, SensorInfoResponse, SetupResponse


class ExplorationProtocolError(Exception):
    pass


class ExplorationProtocol(CommunicationProtocol[SessionConfig]):
    PRF_MAPPING = {
        PRF.PRF_19_5_MHz: "19_5_MHz",
        PRF.PRF_15_6_MHz: "15_6_MHz",
        PRF.PRF_13_0_MHz: "13_0_MHz",
        PRF.PRF_8_7_MHz: "8_7_MHz",
        PRF.PRF_6_5_MHz: "6_5_MHz",
        PRF.PRF_5_2_MHz: "5_2_MHz",
    }
    IDLE_STATE_MAPPING = {
        IdleState.DEEP_SLEEP: "deep_sleep",
        IdleState.SLEEP: "sleep",
        IdleState.READY: "ready",
    }

    @classmethod
    def parse_message(cls, header: dict[str, Any], payload: bytes) -> Message:
        PARSERS = [
            EmptyResultMessage.parse,
            messages.SetBaudrateResponse.parse,
            messages.ErroneousMessage.parse,
            messages.LogMessage.parse,
            ResultMessage.parse,
            messages.SystemInfoResponse.parse,
            SensorInfoResponse.parse,
            SetupResponse.parse,
            messages.StartStreamingResponse.parse,
            messages.StopStreamingResponse.parse,
        ]

        for parser in PARSERS:
            try:
                response = parser(header, payload)
            except messages.ParseError:
                pass
            else:
                return response

        msg = f"Could not parse response with header:\n{header}"
        raise RuntimeError(msg)

    @classmethod
    def setup_command(
        cls,
        session_config: SessionConfig,
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> bytes:
        result = cls._setup_command_preprocessing(session_config)
        result["groups"] = cls._translate_groups_representation(result["groups"])
        if calibrations:
            result["calibration_info"] = []
            for sensor_id, calibration in calibrations.items():
                result["calibration_info"].append(
                    {"sensor_id": sensor_id, "data": calibration.data}
                )
        return (
            json.dumps(
                result,
                separators=(",", ":"),
                ensure_ascii=True,
            )
            + "\n"
        ).encode("ascii")

    @classmethod
    def _setup_command_preprocessing(cls, session_config: SessionConfig) -> dict[str, Any]:
        result = session_config.to_dict()

        if session_config.update_rate is None:
            del result["update_rate"]

        # Exploration server is not interested in this.
        result.pop("extended")

        result["groups"] = map_over_extended_structure(cls._translate_prf_enums, result["groups"])
        result["groups"] = map_over_extended_structure(
            cls._translate_profile_enums, result["groups"]
        )
        result["groups"] = map_over_extended_structure(
            cls._translate_idle_state_enums, result["groups"]
        )
        result["groups"] = map_over_extended_structure(
            cls._translate_sentiel_values, result["groups"]
        )

        result["cmd"] = "setup"
        return result

    @classmethod
    def _translate_profile_enums(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        for subsweep_config_dict in sensor_config_dict["subsweeps"]:
            subsweep_config_dict["profile"] = subsweep_config_dict["profile"].value
        return sensor_config_dict

    @classmethod
    def _translate_prf_enums(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        for subsweep_config_dict in sensor_config_dict["subsweeps"]:
            subsweep_config_dict["prf"] = cls.PRF_MAPPING[subsweep_config_dict["prf"]]
        return sensor_config_dict

    @classmethod
    def _translate_idle_state_enums(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        sensor_config_dict["inter_frame_idle_state"] = cls.IDLE_STATE_MAPPING[
            sensor_config_dict["inter_frame_idle_state"]
        ]
        sensor_config_dict["inter_sweep_idle_state"] = cls.IDLE_STATE_MAPPING[
            sensor_config_dict["inter_sweep_idle_state"]
        ]
        return sensor_config_dict

    @classmethod
    def _translate_sentiel_values(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        """'sweep_rate' and 'frame_rate' has unlocked rate when they are = None.
        RSS's sentinel is 0.0 (zero)
        """
        sensor_config_dict["sweep_rate"] = sensor_config_dict["sweep_rate"] or 0.0
        sensor_config_dict["frame_rate"] = sensor_config_dict["frame_rate"] or 0.0
        return sensor_config_dict

    @classmethod
    def _translate_groups_representation(
        cls, groups_list: list[dict[int, Any]]
    ) -> list[list[dict[str, Any]]]:
        """
        This function translates the Exptool representation, which is

        groups = [
            {sensor_id1: config1, sensor_id2: config2, ...},  # Group 1
            ...,
        ]

        To the representation the Exploration server expects;

        groups = [
            [  # Group 1
                {"sensor_id": sensor_id1, "config": config1},
                {"sensor_id": sensor_id2, "config": config2},
                ...
            ],
            ...
        ]

        """
        return [
            [{"sensor_id": sensor_id, "config": config} for sensor_id, config in group.items()]
            for group in groups_list
        ]



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import abc
import typing as t

import typing_extensions as te

from acconeer.exptool._core.entities import ClientInfo


_ConfigT = t.TypeVar("_ConfigT")
_MetadataT = t.TypeVar("_MetadataT")
_ResultT = t.TypeVar("_ResultT")
_ServerInfoT = t.TypeVar("_ServerInfoT")


_RecorderT = t.TypeVar("_RecorderT")


class ClientError(Exception):
    pass


class ServerError(Exception):
    pass


class ClientCreationError(Exception):
    pass


class ClientABCWithGoodError(abc.ABC):
    def __new__(cls, *args: t.Any, **kwargs: t.Any) -> ClientABCWithGoodError:
        try:
            return super().__new__(cls)
        except TypeError:  # te here is the "Can't instantiate ..."-error
            if cls == Client:
                msg = "Client cannot be instantiated, use Client.open()"
                raise ClientCreationError(msg) from None
            else:
                raise


class Client(
    ClientABCWithGoodError,
    t.Generic[_ConfigT, _MetadataT, _ResultT, _ServerInfoT, _RecorderT],
):
    __registry: te.Final[list[type]] = []

    def __init__(self, client_info: ClientInfo) -> None:
        self._client_info = client_info

        self._session_is_started = False

        self._recorder: t.Optional[_RecorderT] = None
        self._metadata: t.Optional[_MetadataT] = None

    @classmethod
    def open(
        cls,
        ip_address: t.Optional[str] = None,
        tcp_port: t.Optional[int] = None,
        serial_port: t.Optional[str] = None,
        usb_device: t.Optional[t.Union[str, bool]] = None,
        mock: t.Optional[bool] = None,
        override_baudrate: t.Optional[int] = None,
        flow_control: bool = True,
        generation: t.Optional[str] = None,
    ) -> te.Self:
        """
        Open a new client
        """
        if len([e for e in [ip_address, serial_port, usb_device, mock] if e is not None]) > 1:
            msg = "Only one connection can be selected"
            raise ValueError(msg)

        for subclass in cls.__registry:
            try:
                # For a class to be in the "__registry"-list it needs to be a subclass,
                # which it is since it's added in "__init_subclass__".
                #
                # It will also always have the ".open" classmethod, thanks to MRO.
                #
                # That is why these errors are ignored instead of handled.
                return subclass.open(  # type: ignore[no-any-return, attr-defined]
                    ip_address,
                    tcp_port,
                    serial_port,
                    usb_device,
                    mock,
                    override_baudrate,
                    flow_control,
                    generation,
                )
            except ClientCreationError:
                continue

        # This should not happen since the current implementation
        # only has two clients which are mutual exclusive.
        # * The mock client (mock is not None)
        # * The exploration client (mock is None)
        raise ClientCreationError(
            "No client could be created"
            + (". Try specifying 'generation'" if generation is None else "")
        )

    @classmethod
    def __init_subclass__(cls, *, register: bool, **kwargs: t.Any) -> None:
        """
        Registers a subclass if register == True

        Subclasses specifies whether they should be registered in the
        "inherintance list"; class ClientSubclass(Client, register=True)
        """
        super.__init_subclass__(**kwargs)
        if register:
            cls.__registry.append(cls)

    def attach_recorder(self, recorder: _RecorderT) -> None:
        if self.session_is_started:
            msg = "Cannot attach a recorder when session is started."
            raise ClientError(msg)

        if not self.connected:
            msg = "Cannot attach a recorder to a closed client"
            raise ClientError(msg)

        if self._recorder is not None:
            raise ClientError(
                "Client already has a recorder attached. "
                + "Try detaching the current recorder before attaching a new recorder."
            )

        self._recorder = recorder
        self._recorder_start(recorder)

    @abc.abstractmethod
    def _recorder_start(self, recorder: _RecorderT) -> None: ...

    @abc.abstractmethod
    def _recorder_start_session(self) -> None: ...

    @abc.abstractmethod
    def _recorder_sample(self, result: _ResultT) -> None: ...

    @abc.abstractmethod
    def _recorder_stop_session(self) -> None: ...

    def detach_recorder(self) -> t.Optional[_RecorderT]:
        if self.session_is_started:
            msg = "Cannot detach a recorder when session is started."
            raise ClientError(msg)

        if not self.connected:
            msg = "Cannot detach a recorder from a closed client"
            raise ClientError(msg)

        if self._recorder is None:
            return None
        else:
            previously_attached_recorder = self._recorder
            self._recorder = None
            return previously_attached_recorder

    @abc.abstractmethod
    def setup_session(self, config: _ConfigT) -> _MetadataT: ...

    @abc.abstractmethod
    def start_session(self) -> None:
        """Starts the already set up session.

        After this call, the server starts streaming data to the client.

        :raises: ``ClientError`` if ``Client``'s  session is not set up.
        """
        ...

    @abc.abstractmethod
    def get_next(self) -> _ResultT: ...

    @abc.abstractmethod
    def stop_session(self) -> None:
        """Stops an on-going session

        :raises:
            ``ClientError`` if ``Client``'s session is not started.
        """
        ...

    @abc.abstractmethod
    def close(self) -> None:
        """Closes the connection to the host"""
        ...

    @property
    @abc.abstractmethod
    def connected(self) -> bool:
        """Whether this Client is connected."""
        ...

    @property
    @abc.abstractmethod
    def server_info(self) -> _ServerInfoT:
        """The ``ServerInfo``."""
        ...

    @property
    def session_is_setup(self) -> bool:
        """Whether this Client has a session set up."""
        return self._metadata is not None

    @property
    def session_is_started(self) -> bool:
        """Whether this Client's session is started."""
        return self._session_is_started

    @property
    def client_info(self) -> ClientInfo:
        """The ``ClientInfo``."""
        return self._client_info

    def _assert_connected(self) -> None:
        if not self.connected:
            msg = "Client is not connected."
            raise ClientError(msg)

    def _assert_session_setup(self) -> None:
        self._assert_connected()
        if not self.session_is_setup:
            msg = "Session is not set up."
            raise ClientError(msg)

    def _assert_session_started(self) -> None:
        self._assert_session_setup()
        if not self.session_is_started:
            msg = "Session is not started."
            raise ClientError(msg)

    def __enter__(self) -> te.Self:
        return self

    def __exit__(self, *_: t.Any) -> None:
        self.close()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved
from __future__ import annotations

import typing as t

import attrs
import typing_extensions as te

from .message import Message, ParseError


class SystemInfoDict(te.TypedDict):
    rss_version: str
    sensor: str
    sensor_count: int
    ticks_per_second: int
    hw: te.NotRequired[t.Optional[str]]
    max_baudrate: te.NotRequired[int]


@attrs.frozen
class SystemInfoResponse(Message):
    system_info: SystemInfoDict

    @classmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> SystemInfoResponse:
        try:
            return cls(header["system_info"])
        except KeyError as ke:
            raise ParseError from ke



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import enum
import functools
import typing as t

from PySide6.QtCore import Qt
from PySide6.QtGui import QFontMetricsF
from PySide6.QtWidgets import QGridLayout, QLabel, QPushButton, QTextEdit, QWidget

from .syntax_highlight import SyntaxHighlighter


class PresentationType(enum.Enum):
    """
    Enum describing a presentation type.

    A PresenterFunc accept an PresentationType instance as its second argument.
    """

    JSON = "JSON"
    C_SET_CONFIG = "C set_config"

    @property
    def button_label(self) -> str:
        return self.value


PresenterFunc = t.Callable[[t.Any, PresentationType], t.Optional[str]]


class PresentationWindow(QWidget):
    _NOTE_TEXT = "<b>Note!</b> A compact JSON string is always saved to file."
    _HIGHLIGHTERS = {
        PresentationType.JSON: SyntaxHighlighter.json,
        PresentationType.C_SET_CONFIG: SyntaxHighlighter.rss_c,
    }

    def __init__(
        self, presentations: dict[PresentationType, str], parent: t.Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)

        self._preview = QTextEdit(parent=self)
        self._preview.setFontFamily("monospace")
        self._preview.setReadOnly(True)
        self._preview.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)
        self._highlighter = SyntaxHighlighter.json(self._preview.document())

        tab_stop_distance = QFontMetricsF(self._preview.font()).horizontalAdvance(" ") * 8
        self._preview.setTabStopDistance(tab_stop_distance)
        self._preview.setText(
            presentations.get(PresentationType.JSON, "No JSON presentation available")
        )

        layout = QGridLayout()

        num_columns = len(PresentationType)
        layout.addWidget(QLabel(f"<i>{self._NOTE_TEXT}</i>"), 0, 0, 1, num_columns)
        layout.addWidget(self._preview, 1, 0, 1, num_columns)
        for idx, member in enumerate(PresentationType):
            presentation_exists = member in presentations
            button_label = member.button_label + (
                "" if presentation_exists else " (Not available)"
            )

            button = QPushButton(button_label)
            button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            button.setEnabled(presentation_exists)
            button.clicked.connect(
                functools.partial(self._preview.setText, presentations.get(member, ""))
            )
            button.clicked.connect(functools.partial(self._swap_highlighter, member))
            layout.addWidget(button, 2, idx)

        self.setLayout(layout)

    def _swap_highlighter(self, presentation_type: PresentationType) -> None:
        self._highlighter.deleteLater()
        self._highlighter = self._HIGHLIGHTERS[presentation_type](self._preview.document())



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import abc
import typing as t

import typing_extensions as te

from acconeer.exptool._core.communication import Client as BaseClient
from acconeer.exptool._core.communication import ClientCreationError, ClientError
from acconeer.exptool._core.entities import ClientInfo
from acconeer.exptool.a121._core.entities import (
    Metadata,
    Result,
    SensorCalibration,
    SensorConfig,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.recording import Recorder
from acconeer.exptool.a121._core.utils import unextend


class Client(
    BaseClient[
        SessionConfig,  # Config type
        t.List[t.Dict[int, Metadata]],  # Metadata type
        t.List[t.Dict[int, Result]],  # Result type
        ServerInfo,  # Server info type
        Recorder,  # Recorder type
    ],
    register=False,
):
    @classmethod
    def open(
        cls,
        ip_address: t.Optional[str] = None,
        tcp_port: t.Optional[int] = None,
        serial_port: t.Optional[str] = None,
        usb_device: t.Optional[t.Union[str, bool]] = None,
        mock: t.Optional[bool] = None,
        override_baudrate: t.Optional[int] = None,
        flow_control: bool = True,
        generation: t.Optional[str] = "a121",
    ) -> te.Self:
        if generation != "a121":
            raise ClientCreationError

        return super().open(
            ip_address,
            tcp_port,
            serial_port,
            usb_device,
            mock,
            override_baudrate,
            flow_control,
            generation="a121",
        )

    def __init__(self, client_info: ClientInfo) -> None:
        super().__init__(client_info)
        self._sensor_calibrations: t.Optional[dict[int, SensorCalibration]] = None
        self._calibrations_provided: dict[int, bool] = {}
        self._session_config: t.Optional[SessionConfig] = None

    def _return_results(
        self, extended_results: list[dict[int, Result]]
    ) -> t.Union[Result, list[dict[int, Result]]]:
        if self.session_config.extended:
            return extended_results
        else:
            return unextend(extended_results)

    @property
    def session_config(self) -> SessionConfig:
        """The :class:`SessionConfig` for the current session"""
        self._assert_session_setup()
        assert self._session_config is not None  # Should never happen if session is setup
        return self._session_config

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        """The extended :class:`Metadata` for the current session"""
        self._assert_session_setup()
        assert self._metadata is not None  # Should never happen if session is setup
        return self._metadata

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        """
        Returns a dict with a :class:`SensorCalibration` per used
        sensor for the current session:

        For example, if session_setup was called with

        .. code-block:: python

            client.setup_session(
                SessionConfig({1: SensorConfig(), 3: SensorConfig()}),
            )

        this attribute will return {1: SensorCalibration(...), 3: SensorCalibration(...)}
        """
        self._assert_session_setup()

        if not self._sensor_calibrations:
            msg = "Server did not provide calibration"
            raise ClientError(msg)

        return self._sensor_calibrations

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        """
        Returns whether a calibration was provided for each sensor in
        setup_session. For example, if setup_session was called with

        .. code-block:: python

            client.setup_session(
                SessionConfig({1: SensorConfig(), 2: SensorConfig()}),
                calibrations={2: SensorCalibration(...)},
            )

        this attribute will return ``{1: False, 2: True}``
        """
        return self._calibrations_provided

    @abc.abstractmethod
    def setup_session(  # type: ignore[override]
        self,
        config: t.Union[SensorConfig, SessionConfig],
        calibrations: t.Optional[dict[int, SensorCalibration]] = None,
    ) -> t.Union[Metadata, list[dict[int, Metadata]]]:
        """Sets up the session specified by ``config``.

        :param config: The session to set up.
        :param calibrations: An optional dict with :class:`SensorCalibration` for the session.
        :raises:
            ``ValueError`` if the config is invalid.

        :returns:
            ``Metadata`` if ``config.extended is False``,
            ``list[dict[int, Metadata]]`` otherwise.
        """
        ...

    @abc.abstractmethod
    def get_next(self) -> t.Union[Result, list[dict[int, Result]]]:  # type: ignore[override]
        """Gets results from the server.

        :returns:
            A ``Result`` if the setup ``SessionConfig.extended is False``,
            ``list[dict[int, Result]]`` otherwise.
        :raises:
            ``ClientError`` if ``Client``'s session is not started.
        """
        ...

    def _recorder_start(self, recorder: Recorder) -> None:
        recorder._start(
            client_info=self.client_info,
            server_info=self.server_info,
        )

    def _recorder_start_session(self) -> None:
        if self._recorder is not None:
            calibrations_provided: t.Optional[dict[int, bool]] = self.calibrations_provided
            try:
                calibrations = self.calibrations
            except ClientError:
                calibrations = None
                calibrations_provided = None

            self._recorder._start_session(
                config=self.session_config,
                metadata=self.extended_metadata,
                calibrations=calibrations,
                calibrations_provided=calibrations_provided,
            )

    def _recorder_sample(self, result: list[dict[int, Result]]) -> None:
        if self._recorder is not None:
            self._recorder._sample(result)

    def _recorder_stop_session(self) -> None:
        if self._recorder is not None:
            self._recorder._stop_session()

    def __enter__(self) -> Client:
        return self

    def __exit__(self, *_: t.Any) -> None:
        self.close()



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from ._processor import (
    AmplitudeMethod,
    Processor,
    ProcessorConfig,
    ProcessorResult,
    SubsweepProcessorResult,
    get_sensor_config,
)
